#!/usr/bin/perl
# Top-level script for creating a skel that contains the infrastructure
# for all virtual servers.
# Copyright (c) 1999, 2000 Designer Servers Ltd.
# Contributed by Nick Burrett <nick@dsvr.net>
#
# This file is part of the Virtual Server Administrator (FreeVSD)
#
# FreeVSD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# FreeVSD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FreeVSD; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

# command syntax:
#   create-skel.pl <mnt point>
#
#   mntpoint - mount point of the disk to contain virtual servers
#
use Sys::Hostname;
use IO::File;
use File::Copy;
use File::Find;
use File::stat;

sub read_config {
  open (CONFIG, "/etc/freevsd.conf")
    or die "Configuration file /etc/freevsd.conf is required";
  while (<CONFIG>) {
    chomp; # remove newlines
    s/\#.*//; # remove comments
    s/^\s+//; # no leading white
    s/\s+$//; # no trailing white
    next unless length; # skip blank lines
    my ($var, $value) = split (/\s*=\s*/, $_, 2);
    $prefs{$var} = $value;
  }
  close (CONFIG);
}

sub skelcopy {
    my ($temp, $temp1, $mode);
    $temp = $File::Find::name;
    ($temp1 = $temp) =~ s/$skelsrc\///;
    if (-d $temp && $temp !~ /CVS/) {
        mkdir ("$skelroot/$temp1", 0755);
    } elsif ($temp !~ /CVS/) {
	$mode = (stat ($temp))[2];
	`cp -a $temp $skelroot/$temp1`;
	chown 0, 0, "$skelroot/$temp1";
    }
}

# Check parameters
if (@ARGV != 1) {
    printf ("syntax: create-skel.pl <mount point>\n");
    exit 1;
}

$mntpoint = $ARGV[0];
unless (-d "$mntpoint") { die "$mntpoint does not exist\n"; }
$skelroot = $mntpoint."/skel";
read_config ();

# Install various FreeVSD files
print "-- creating skel\n";
if (-d "$skelroot") {
    print "-- deleting old skel\n";
    `rm -rf $skelroot`;
}
mkdir ("$skelroot", 0755);
foreach $dir ("bin", "contrib", "dev", "etc", "home", "lib", "proc", "tmp",
	      "usr", "sbin") {
    mkdir ("$skelroot/$dir", 0755);
}

print "-- merging skel repositories\n";

if (defined ($skelsrc = $prefs{"generic_skelrepo"})) {
  if (-d "$skelsrc") {
    find \&skelcopy, $skelsrc;
  } else {
    die "$skelsrc does not exist";
  }
} else {
  die "/etc/freevsd.conf needs a `generic_skelrepo' definition";
}

if (defined ($skelsrc = $prefs{"os_skelrepo"})) {
  if (-d "$skelsrc") {
    find \&skelcopy, $skelsrc;
  } else {
    die "$skelsrc does not exist";
  }
} else {
  die "/etc/freevsd.conf needs a `os_skelrepo' definition";
}

if (defined ($skelsrc = $prefs{"site_skelrepo"})) {
  if (-d "$skelsrc") {
    find \&skelcopy, $skelsrc;
  } else {
    die "$skelsrc does not exist";
  }
}

if (! defined ($dist_root = $prefs{"dist_root"})) {
    die "/etc/freevsd.conf requires a `dist_root' definition";
}
$src = $dist_root."/src";
if (! -d "$src/vsd") {
    die "directive `dist_root' does not appear to point to a valid FreeVSD source directory structure";
}

print "-- merging /bin programs\n";
@list = split (/ /, $prefs{"bin_progs"});
foreach $file (@list) {
    `cp -a /bin/$file $skelroot/bin/$file`;
}

print "-- merging /sbin programs\n";
@list = split (/ /, $prefs{"sbin_progs"});
foreach $file (@list) {
    `cp -a /sbin/$file $skelroot/sbin/$file`;
}

print "-- merging /lib\n";
`cp -a /lib $skelroot/`;

print "-- merging /usr\n";
`cp -a /usr $skelroot/`;

print "-- merging /usr/sbin programs\n";
`rm -r -f $skelroot/usr/sbin`;
mkdir ("$skelroot/usr/sbin", 0755);
@list = split (/ /, $prefs{"usr_sbin_progs"});
foreach $file (@list) {
    `cp -a /usr/sbin/$file $skelroot/usr/sbin/$file`;
}

print "-- merging /usr/libexec programs\n";
`rm -r -f $skelroot/usr/libexec`;
mkdir ("$skelroot/usr/libexec", 0755);
@list = split (/ /, $prefs{"usr_libexec_progs"});
foreach $file (@list) {
    `cp -a /usr/libexec/$file $skelroot/usr/libexec/$file`;
}

print "-- deleting extra files\n";
@list = split (/ /, $prefs{"delete_progs"});
foreach $file (@list) {
    system ("rm -r -f $skelroot/$file");
}

print "-- merging skel repositories\n";
if (defined ($skelsrc = $prefs{"generic_skelrepo"})) {
  if (-d "$skelsrc") {
    find \&skelcopy, $skelsrc;
  } else {
    die "$skelsrc does not exist";
  }
} else {
  die "/etc/freevsd.conf needs a `generic_skelrepo' definition";
}

if (defined ($skelsrc = $prefs{"os_skelrepo"})) {
  if (-d "$skelsrc") {
    find \&skelcopy, $skelsrc;
  } else {
    die "$skelsrc does not exist";
  }
} else {
  die "/etc/freevsd.conf needs a `os_skelrepo' definition";
}

if (defined ($skelsrc = $prefs{"site_skelrepo"})) {
  if (-d "$skelsrc") {
    find \&skelcopy, $skelsrc;
  } else {
    die "$skelsrc does not exist";
  }
}

# install (mode, source filename, destination filename)
sub install {
    my ($mode, $src, $dst);
    $mode = $_[0];
    $src = $_[1];
    $dst = $_[2];

    `install -s -g root -o root -m $mode $src $dst`;
}

# Install various FreeVSD files
install ("0755", "$src/vsd/vsclients", "$skelroot/usr/sbin");
`ln -s vsclients $skelroot/usr/sbin/quotastats`;
`ln -s vsclients $skelroot/usr/sbin/setquota`;
`ln -s vsclients $skelroot/usr/sbin/rebootvs`;

install ("4755", "$src/userutil/useradd", "$skelroot/usr/sbin/useradd");
install ("4755", "$src/userutil/userdel", "$skelroot/usr/sbin/userdel");
install ("4755", "$src/userutil/usermod", "$skelroot/usr/sbin/usermod");
install ("4755", "$src/userutil/groupadd", "$skelroot/usr/sbin/groupadd");
install ("4755", "$src/userutil/groupdel", "$skelroot/usr/sbin/groupdel");
install ("4755", "$src/userutil/groupmod", "$skelroot/usr/sbin/groupmod");
install ("4755", "$src/userutil/passwd", "$skelroot/usr/bin/passwd");
install ("0755", "$src/userutil/listrights", "$skelroot/usr/sbin/listrights");
install ("4755", "$src/userutil/setrights", "$skelroot/usr/sbin/setrights");

# Install login
install ("0755", "$src/login/login", "$skelroot/bin/login");

print "-- installing packages in freevsd/src/extras\n";
# Install the additional applications in freevsd/src/extras
# Our method move the files in freevsd/src/extras into the skel
# and then perform the installation from there.
if (! defined ($childpid = fork ())) {
    die "cannot fork: $!";
} elsif ($childpid == 0) {
    # child process
    chdir ("$dist_root");
    `tar -cf $skelroot/tmp/extras.tar ./`;
    chroot ($skelroot);
    chdir ("/tmp");
    `tar -xf extras.tar`;
    unlink ("extras.tar");
    @dirs = glob ("/tmp/src/extras/*");
    foreach (@dirs) {
	chdir ($_);
	`make install`;
    }
    chdir ("/");
    `rm -r -f /tmp`;
    mkdir ("/tmp", 01777);
    exit;
} else {
    # Parent process.
    waitpid ($childpid, 0);
}

# Do operating system specific skel installation
if (defined ($pscript = $prefs{"os_skel_install"})) {
  if (-x "$pscript") {
      do "$pscript";
  }
} else {
  die "/etc/freevsd.conf needs a `os_skel_install' definition";
}

# Do site specific skel installation
if (defined ($pscript = $prefs{"site_skel_install"})) {
  if (-x "$pscript") {
      do "$pscript";
  }
}

