#!/usr/bin/perl
# Top-level script for creating virtual servers from a skeleton template
# Copyright (c) 1999, 2000 Designer Servers Ltd.
# Contributed by Nick Burrett <nick@dsvr.net>
#
# This file is part of the Virtual Server Administrator (FreeVSD)
#
# FreeVSD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# FreeVSD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FreeVSD; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

# To run this script, you must have previously created a skel with
# create-skel.pl.
#
# This script could be called as a cron job, periodically process
# /tmp/vsd.accounts creating and deleting accounts whenever necessary.
#
use Sys::Hostname;
use IO::File;
use File::Copy;

# /tmp/vsd.accounts is a temporary file that contains the queue of
# virtual servers that need creating.
unless (-e "/tmp/vsd.accounts") { exit 0; }

# If this lock file exists then account creation is already in
# progress, don't start another job.
if (-e "/tmp/vsd.accounts.lock") { exit 0; }
rename ("/tmp/vsd.accounts", "/tmp/vsd.accounts.bak");

$sendmail="/usr/sbin/sendmail"; # Location of sendmail
$hostname = hostname ();
srand (time ^ $$);

read_vsd_conf ();
read_config ();

if (! defined ($prefs{"os_vs_install"})) {
    die "/etc/freevsd.conf needs a `os_vs_install' definition";
}
require ($prefs{"os_vs_install"});

if (defined ($prefs{"site_vs_install"})) {
  # Purely optional.
  require ($prefs{"site_vs_install"});
}

# Array of characters suitable for an account password.
@chars = ("A".."Z", "a".."z", 0..9);

open (JOB, "/tmp/vsd.accounts.bak");
while (<JOB>) {
  my ($action, $vs, $param) = split (/ /, $_);
  chomp $param;
  if ($action eq "create") {
    $fqdn = $param;

    if (! -d $conf_mount{$conf_partition{$vs}}) {
      printf ("Couldn't process %s: mount point `%s' does not exist\n", $name, $conf_mount{$conf_partition{$vs}});
    } else {
      # Create a random password for the account
      $password = join ("", @chars[map {rand @chars} (1..8)]);
      $mntpoint = $conf_mount{$conf_partition{$vs}};
      $skelroot = $mntpoint."/skel"; # Used in all sub-procs
      $vsdir = $mntpoint."/vs/".$vs; # Used in all sub-procs

      make_vs ($vs);
      os_make_vs ($vs);
      if (defined ($prefs{"site_vs_install"})) {
	  site_make_vs ($vs);
      }

      config_vs ($vs, $conf_ip{$vs}, $fqdn, $conf_startuid{$vs}, $conf_users{$vs});
      os_config_vs ($vs, $conf_ip{$vs}, $fqdn, $conf_startuid{$vs}, $conf_users{$vs});

      if (defined ($prefs{"site_vs_install"})) {
	  site_config_vs ($vs, $conf_ip{$vs}, $fqdn, $conf_startuid{$vs}, $conf_users{$vs});
      }

      if ($prefs{"send_mail"} eq "YES") {
	email ($fqdn, $conf_ip{$vs}, $vs, $password, $prefs{"mail_template"});
      }
    }
  }
  if ($action eq "delete") {
    my ($mount) = $param;
    if (! -d "$mount") {
      printf ("Couldn't delete %s: mount point '%s' does not exist\n", $name, $mount);
    } else {
      `rm -r -f $mount/vs/$vs`;
    }
  }
}
close (JOB);
unlink ("/tmp/vsd.accounts.bak");
unlink ("/tmp/vsd.accounts.lock");
exit 0;

sub read_config {
  open (CONFIG, "/etc/freevsd.conf")
    or die "Configuration file /etc/freevsd.conf is required";
  while (<CONFIG>) {
    chomp; # remove newlines
    s/\#.*//; # remove comments
    s/^\s+//; # no leading white
    s/\s+$//; # no trailing white
    next unless length; # skip blank lines
    my ($var, $value) = split (/\s*=\s*/, $_, 2);
    $prefs{$var} = $value;
  }
  close (CONFIG);
}

# The following arrays are created from /etc/vsd.conf
#   mount
#   device
#   server
#   ip
#   startuid
#   users
#   status
#   quota
#   partition

sub read_vsd_conf {
    my ($p1, $p2, $p3, $sname);
    open (CONFIG, "/etc/vsd.conf")
	or die "Configuration file /etc/vsd.conf is required";
    while (<CONFIG>) {
	chomp; # remove newlines
	s/\#.*//; # remove comments
	s/^\s+//; # no leading white
	s/\s+$//; # no trailing white
	next unless length; # skip blank lines
	if (/^mount/) {
	    ($p1, $p2, $p3) = split (/ /, $_);
	    $conf_mount{$p2} = $p3;
	}
	if (/^device/) {
	    ($p1, $p2, $p3) = split (/ /, $_);
	    $conf_device{$p2} = $p3;
	}
	if (/^server/) { # server decl
	    ($p1, $sname) = split (/ /, $_);
	    push (@conf_server, $sname);
	}
	if (/^ip/) {  # ip param in server decl
	    ($p1, $conf_ip{$sname}) = split (/ /, $_);
	}
	if (/^startuid/) {  # ip param in server decl
	    ($p1, $conf_startuid{$sname}) = split (/ /, $_);
	}
	if (/^users/) {  # users param in server decl
	    ($p1, $conf_users{$sname}) = split (/ /, $_);
	}
	if (/^status/) {  # status param in server decl
	    ($p1, $conf_status{$sname}) = split (/ /, $_);
	}
	if (/^quota/) {  # quota param in server decl
	    ($p1, $conf_quota{$sname}) = split (/ /, $_);
	}
	if (/^partition /) {  # partition param in server decl
	    ($p1, $conf_partition{$sname}) = split (/ /, $_);
	}
    }
    close (CONFIG);
}

# make_vs <virtual server name>
sub make_vs {
  my ($vs);

  $vs = $_[0];

  unless (-d "$mntpoint/vs") {
    mkdir ("$mntpoint/vs", 0755);
  }
  unless (-d "$vsdir") {
    mkdir ("$vsdir", 0755);
  }
  # Files are copied as hard links
  system("cp -al $skelroot/* $vsdir");
  system("rm -r -f $vsdir/etc");
  system("cp -a $skelroot/etc $vsdir");
  mkdir ("$vsdir/etc/vsd", 0755);
  # Some versions of crond require a /etc/cron.d directory.
  mkdir ("$vsdir/etc/cron.d", 0755);
  mkdir ("$vsdir/dev", 0755);
  mkdir ("$vsdir/root", 0755);
  mkdir ("$vsdir/var", 0755);
}

# config_vs <server> <ipaddr> <fqdn> <startuid> <numuid>
sub config_vs {
  my ($vs, $ip, $fqdn, $startuid, $numuid, $enduid, $c, $fh);

  $vs = $_[0];
  $ip = $_[1];
  $fqdn = $_[2];
  $startuid = $_[3];
  $numuid = $_[4];
  $enduid = $startuid + $numuid - 1;

  # This file is used by the uname system call
  open (OUT, ">$vsdir/etc/FQDN");
  printf OUT "%s %s\n", $fqdn, $vs;
  close (OUT);

  # Create /etc/hosts
  open (OUT, ">$vsdir/etc/hosts");
  printf OUT "%s %s localhost\n", $ip, $fqdn;
  close (OUT);

  # Create /etc/resolv.conf
  open (OUT, ">$vsdir/etc/resolv.conf");
  printf OUT "domain %s\n", $fqdn;
  printf OUT "nameserver %s\n", $prefs{"primary_ns"};
  printf OUT "nameserver %s\n", $prefs{"secondary_ns"};
  close (OUT);

  # Create /etc/issue.net
  open (OUT, ">$vsdir/etc/issue.net");
  printf OUT "Server %s\n", $fqdn;
  close (OUT);

  # Create /etc/sendmail.cw
  open (OUT, ">$vsdir/etc/sendmail.cw");
  printf OUT "# sendmail.cw - include all aliases for your machine here\n";
  printf OUT "%s\n", $fqdn;
  close (OUT);

  `cp $vsdir/etc/login.defs $vsdir/tmp`;
  open (IN, "$vsdir/tmp/login.defs");
  open (OUT, ">$vsdir/etc/login.defs");
  while (<IN>) {
    $_ =~ s/\#UID_MIN\#/$startuid/;
    $_ =~ s/\#UID_MAX\#/$enduid/;
    $_ =~ s/\#GID_MIN\#/$startuid/;
    $_ =~ s/\#GID_MAX\#/$enduid/;
    printf OUT "%s", $_;
  }
  close (OUT);
  close (IN);
  unlink ("$vsdir/tmp/login.defs");

  if (!defined($childpid = fork ())) {
    die "cannot fork: $!";
  } elsif ($childpid == 0) {
    # child
      my ($gid_admin, $gid_mail, $uid_admin, $uid_mail);
    chroot ($vsdir);
    chdir ("/");
    $gid_admin = $startuid;
    $gid_mail = $startuid + 1;
    $uid_admin = $gid_admin;
    $uid_mail = $gid_mail;
    
    `/usr/sbin/groupadd admin -g $gid_admin`;
    `/usr/sbin/groupadd mail -g $gid_mail`;
    `/usr/sbin/useradd -c administrator -d /root -u $uid_admin -g admin -G web -M -n -s /bin/bash admin`;
    `/usr/sbin/useradd -c mail -d /var/spool/mail -u $uid_mail -g mail -M -n -s /bin/false mail`;
    exit;
  } else {
    waitpid ($childpid, 0);
  }
}

# email <fqdn> <ipaddr> <server> <admin-password>
sub email {
  my ($fqdn, $ipaddr, $vs, $pw, $email, $tmpl);

  $fqdn = $_[0];
  $ipaddr = $_[1];
  $vs = $_[2];
  $pw = $_[3];
  $email = $prefs{"mail_addr"};
  $tmpl = $_[4];
  
  open (TMPL, "< $tmpl")
    or warn "cannot open $tmpl: $!\n";
  open (SMAIL, "|$sendmail -oi -t")
    or die "cannot open a pipe to $sendmail: $!";
  printf (SMAIL "From: vsd@%s (Virtual Server Administrator)\n", $hostname);
  printf (SMAIL "To: %s\n", $email);
  while (<TMPL>) {
    s/\#VS\#/$vs/;
    s/\#HOSTNAME\#/$fqdn/;
    s/\#IPADDRESS\#/$ipaddr/;
    s/\#PASSWORD\#/$pw/;
    printf SMAIL "%s", $_;
  }
  close (SMAIL);
  close (TMPL);
}
