/* Obtain information from the configuration files.
   Copyright (c) 1999 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "libvsd.h"

#define VSD_CONF "/etc/vsd.conf"

#define MIN_UID 1000
#define MAX_UID 65000

/* Virtual servers have uid's allocated in minimum chunks of 100.  */
#define CHUNK 100

struct keyword
{
  const char *name;
  int token;
  int arguments;
};

#define T_DEVICE 0
#define T_IP 1
#define T_IPALIAS 2
#define T_MAXVS 3
#define T_MOUNT 4
#define T_PARTITIONS 5
#define T_PARTITION 6
#define T_QUOTA 7
#define T_SERVER 8
#define T_STARTUID 9
#define T_STATUS 10
#define T_USERS 11

static struct keyword keyword_table[] =
{
  { "device", T_DEVICE, 2 },
  { "ip", T_IP, 1 },
  { "ipalias", T_IPALIAS, 1 },
  { "maxvs", T_MAXVS, 1 },
  { "mount", T_MOUNT, 2 },
  { "partitions", T_PARTITIONS, 1 },
  { "partition", T_PARTITION, 1 },
  { "quota", T_QUOTA, 1 },
  { "server", T_SERVER, 1 },
  { "startuid", T_STARTUID, 1 },
  { "status", T_STATUS, 1 },
  { "users", T_USERS, 1 },
  { NULL, 0 }
};

/* Return an index into keyword_table of a matching `key'. If the key
   cannot be found then return -1.  */
static int lookup_keyword (const char *key)
{
  int x;

  for (x = 0; keyword_table[x].name; x++)
    if (strcmp (keyword_table[x].name, key) == 0)
      return x;

  return -1;
}

struct vsd_vs_map *vsd_map_read (void)
{
  struct vsd_file_line *vsconf, *line;
  struct vsd_vs_map *map;
  char *p, **argv, key[256];
  int x, token, argc, part, type;

  if (vsd_load_file (&vsconf, "/", VSD_CONF))
    return NULL;

  type = 0;
  map = (struct vsd_vs_map *) malloc (sizeof (struct vsd_vs_map));
  if (map == NULL)
    return NULL;

  memset (map, 0, sizeof (struct vsd_vs_map));
  for (line = vsconf; line; line = line->next)
    {
      /* Skip white space.  */
      for (p = line->line; isspace (*p); p++)
	;

      /* Ignore comment lines and skip empty lines.  */
      if (*p == '#' || *p == '\0')
	continue;

      for (x = 0; ! isspace (*p); x++, p++)
	key[x] = *p;
      key[x] = '\0';

      token = lookup_keyword (key);
      if (token == -1)
	{
	  /* Unrecognised token.  */
	  errno = ENOENT;
	  vsd_free_file (vsconf);
	  return NULL;
	}

      argv = vsd_argv_parse (' ', line->line, &argc);
      if (argc - 1 != keyword_table[token].arguments)
	{
	  /* Invalid number of arguments.  */
	  errno = EINVAL;
	  free (argv);
	  vsd_free_file (vsconf);
	  return NULL;
	}

      switch (token)
	{
	case T_PARTITIONS:
	  map->partitions = atoi (argv[1]);
	  map->partition = (struct vsd_disk *) malloc (sizeof (struct vsd_disk)
						       * map->partitions);
	  break;

	case T_DEVICE:
	  part = atoi (argv[1]);

	  if (map->partitions <= part)
	    {
	      /* Number of partitions is not defined/incorrect  */
	      errno = ESRCH;
	      return NULL;
	    }
	  map->partition[part].device = strdup (argv[2]);
	  break;

	case T_MAXVS:
	  map->max_servers = atoi (argv[1]);
	  if (map->max_servers < 0)
	    map->max_servers = 256;
	  break;

	case T_MOUNT:
	  part = atoi (argv[1]);

	  if (map->partitions <= part)
	    {
	      /* Number of partitions is not defined/incorrect.  */
	      errno = ESRCH;
	      return NULL;
	    }
	  map->partition[part].pathname = strdup (argv[2]);
	  break;

	case T_SERVER:
	  map->servers ++;
	  if (map->servers == 1)
	    map->server = (struct vsd_vs *) malloc (sizeof (struct vsd_vs));
	  else
	    map->server = (struct vsd_vs *) realloc (map->server,
						     sizeof (struct vsd_vs)
						     * map->servers);

	  strcpy (map->server[map->servers - 1].name, argv[1]);
	  type = 0;  /* for T_IP */
	  break;

	case T_IPALIAS:
	  map->ipaliases ++;
	  if (map->ipaliases == 1)
	    map->ipalias = (struct vsd_ipalias *)
	      malloc (sizeof (struct vsd_ipalias));
	  else
	    map->ipalias = (struct vsd_ipalias *)
	      realloc (map->ipalias,
		       sizeof (struct vsd_ipalias) * map->ipaliases);
	  strcpy (map->ipalias[map->ipaliases - 1].name, argv[1]);
	  type = 1; /* for T_IP */
	  break;

	case T_IP: /* server/ipalias */
	  if (type == 0)
	    strcpy (map->server[map->servers - 1].ip, argv[1]);
	  else
	    strcpy (map->ipalias[map->ipaliases - 1].ip, argv[1]);
	  break;

	case T_USERS: /* server */
	  map->server[map->servers - 1].users = atoi (argv[1]);
	  break;

	case T_STARTUID: /* server */
	  map->server[map->servers - 1].startuid = atoi (argv[1]);
	  break;

	case T_STATUS: /* server */
	  if (type == 0)
	    {
	      if (strcmp (argv[1], "active") == 0)
		map->server[map->servers - 1].status = S_ACTIVE;
	      else if (strcmp (argv[1], "disabled") == 0)
		map->server[map->servers - 1].status = S_DISABLED;
	    }
	  else
	    {
	      if (strcmp (argv[1], "active") == 0)
		map->ipalias[map->ipaliases - 1].status = S_ACTIVE;
	      else if (strcmp (argv[1], "disabled") == 0)
		map->ipalias[map->ipaliases - 1].status = S_DISABLED;
	    }
	  break;

	case T_QUOTA: /* server */
	  map->server[map->servers - 1].quota = atoi (argv[1]);
	  break;

	case T_PARTITION: /* server */
	  map->server[map->servers - 1].partition = atoi (argv[1]);
	  break;
	}
      free (argv);
    }

  vsd_free_file (vsconf);

  return map;
}

int vsd_map_save (struct vsd_vs_map *map)
{
  FILE *stream;
  int x;

  stream = fopen (VSD_CONF, "w");
  if (stream == NULL)
    return 1;

  fprintf (stream, "# /etc/vsd.conf - FreeVSD configuration file\n");
  fprintf (stream, "# This file is automatically generated\n\n");
  fprintf (stream, "# Maximum number of virtual servers\n");
  fprintf (stream, "maxvs %d\n\n", map->max_servers);
  fprintf (stream,
	   "# Number of partitions being used for virtual server accounts\n");
  fprintf (stream, "partitions %d\n\n", map->partitions);
  fprintf (stream, "# Device name and mount point of each partition\n");

  for (x = 0; x < map->partitions; x++)
    {
      fprintf (stream, "device %d %s\n", x, map->partition[x].device);
      fprintf (stream, "mount %d %s\n", x, map->partition[x].pathname);
    }

  fprintf (stream, "\n# Virtual server account details\n");
  for (x = 0; x < map->servers; x++)
    {
      char *status;

      if (map->server[x].name[0] == '\0')
	continue; /* Skip deleted entries.  */

      fprintf (stream, "server %s\n", map->server[x].name);
      fprintf (stream, "   ip %s\n", map->server[x].ip);
      fprintf (stream, "   startuid %d\n", map->server[x].startuid);
      fprintf (stream, "   users %d\n", map->server[x].users);
      switch (map->server[x].status)
	{
	case S_ACTIVE:
	  status = "active";
	  break;
	case S_DISABLED:
	  status = "disabled";
	  break;
	default:
	  status = "disabled";
	  break;
	}
      fprintf (stream, "   status %s\n", status); 
      fprintf (stream, "   quota %d\n", map->server[x].quota);
      fprintf (stream, "   partition %d\n\n", map->server[x].partition);
    }

  for (x = 0; x < map->ipaliases; x++)
    {
      char *status;

      if (map->ipalias[x].name[0] == '\0')
	continue; /* Skip deleted entries.  */

      fprintf (stream, "ipalias %s\n", map->ipalias[x].name);
      fprintf (stream, "   ip %s\n", map->ipalias[x].ip);
      switch (map->server[x].status)
	{
	case S_ACTIVE:
	  status = "active";
	  break;
	case S_DISABLED:
	  status = "disabled";
	  break;
	}
      fprintf (stream, "   status %s\n\n", status);
    }

  fclose (stream);

  return 0;
}

void vsd_map_free (struct vsd_vs_map *map)
{
  int x;

  if (map == NULL)
    return;

  for (x = 0; x < map->partitions; x++)
    {
      free (map->partition[x].device);
      free (map->partition[x].pathname);
    }
  free (map->partition);
  if (map->servers > 1)
    free (map->server);
  if (map->ipaliases > 0)
    free (map->ipalias);
  free (map);
}

static int allocate_partition (struct vsd_vs_map *map)
{
  int x, p, v;
  int *count = (int *) alloca (map->partitions * sizeof (int));

  memset (count, 0, map->partitions * sizeof (int));
  for (x = 0; x < map->servers; x++)
    count[map->server[x].partition]++;

  /* Find the partition with the lowest count.  */
  p = 0;
  v = 32767;
  for (x = 0; x < map->partitions; x++)
    if (count[x] < v)
      {
	v = count[x];
	p = x;
      }

  return p;
}

/* Add a new virtual server to the vsd.conf configuration file.
   `server_name' is the name of the virtual server.
   `ip' is the IP address of the virtual server
   `start_uid' is the start of the uid range allocated to the server
   `num_users' is the number of uids to allocate to the server
   `quota' is the initial disk quota limit imposed on the server.  */
int vsd_map_add_server (struct vsd_vs_map *map,
			const char *server_name, const char *ip,
			int start_uid, int num_users, int quota)
{
  int partition = allocate_partition (map);
  int x;

  if (map->servers >= map->max_servers)
    {
      /* Adding another server would exceed the limits imposed.  */
      errno = EMFILE;
      return 1;
    }

  map->servers ++;
  map->server = (struct vsd_vs *) realloc (map->server,
					   sizeof (struct vsd_vs)
					   * map->servers);
  if (map->server == NULL)
    return 1;

  x = map->servers - 1;

  strcpy (map->server[x].name, server_name);
  strcpy (map->server[x].ip, ip);
  map->server[x].startuid = start_uid;
  map->server[x].users = num_users;
  map->server[x].quota = quota;
  map->server[x].partition = partition;
  map->server[x].status = S_ACTIVE;

  return 0;
}

/* Delete virtual server `server' and any IP aliases associated with it.
   Return 0 on success and 1 if the server doesn't exist.  */
int vsd_map_delete_server (struct vsd_vs_map *map, const char *server)
{
  int x;
  int found = 0;

  /* Delete virtual server.  */
  for (x = 0; x < map->servers; x++)
    if (map->server[x].name[0] != '\0'
	&& strcmp (map->server[x].name, server) == 0)
      {
	map->server[x].name[0] = '\0';
	found = 1;
      }

  /* Delete IP aliases.  */
  if (found)
    for (x = 0; x < map->ipaliases; x++)
      if (map->ipalias[x].name[0] != '\0'
	  && strcmp (map->ipalias[x].name, server) == 0)
	map->ipalias[x].name[0] = '\0';
 
  return 1;
}

struct vsd_vs *vsd_getmapent (struct vsd_vs_map *map,
			      const char *virtual_server,
			      const char *ip_addr)
{
  int x;

  if ((virtual_server == NULL || *virtual_server == '\0')
      && (ip_addr == NULL || *ip_addr == '\0'))
    {
      errno = EINVAL;
      return NULL;
    }

  for (x = 0; x < map->servers; x++)
    {
      if (map->server[x].name[0] == '\0')
	continue; /* Skip deleted server entries.  */

      if ((virtual_server == NULL || *virtual_server == '\0')
	  && strcmp (ip_addr, map->server[x].ip) == 0)
	return &map->server[x];

      if ((ip_addr == NULL || *ip_addr == '\0')
	  && strcmp (virtual_server, map->server[x].name) == 0)
	return &map->server[x];
    }

  /* Check IP aliases.  */
  if (ip_addr != NULL)
    {
      for (x = 0; x < map->ipaliases; x++)
	{
	  if (map->ipalias[x].name[0] == '\0')
	    continue; /* Skip deleted server entries.  */

	  if ((virtual_server == NULL || *virtual_server == '\0')
	      && strcmp (ip_addr, map->ipalias[x].ip) == 0)
	    return vsd_getmapent (map, map->ipalias[x].name, NULL);
	}
    }

  errno = ENOENT;
  return NULL;
}

const char *vsd_map_server_root (struct vsd_vs_map *map,
				 const char *server_name)
{
  struct vsd_vs *vs = vsd_getmapent (map, server_name, NULL);
  static char *temp = NULL;
  int len;

  if (! vs)
    return NULL;

  len = (strlen (map->partition[vs->partition].pathname)
	 + strlen (vs->name) + 6);

  /* Allocate some storage space for the pathname.  */
  if (temp == NULL)
    temp = (char *) malloc (len);
  else
    temp = (char *) realloc (temp, len);

  if (temp != NULL)
    sprintf (temp, "%s/vs/%s",
	     map->partition[vs->partition].pathname, vs->name);

  return temp;
}

int vsd_map_set_status (struct vsd_vs_map *map, const char *server,
			const char *status)
{
  struct vsd_vs *vs = vsd_getmapent (map, server, NULL);
  int x;

  if (vs == NULL)
    return 1;

  if (strcmp (status, "active") == 0)
    vs->status = S_ACTIVE;
  else if (strcmp (status, "disabled") == 0)
    vs->status = S_DISABLED;

  for (x = 0; x < map->ipaliases; x++)
    if (strcmp (map->ipalias[x].name, server) == 0)
      map->ipalias[x].status = vs->status;

  return 0;
}

void vsd_map_uid_range (struct vsd_vs_map *map, const char *server,
			int *uid_min, int *uid_max)
{
  struct vsd_vs *vs = vsd_getmapent (map, server, NULL);

  *uid_min = *uid_max = 0;
  if (vs != NULL)
    {
      *uid_min = vs->startuid;
      *uid_max = *uid_min + vs->users;
    }
}

/* Return -1 on error or a start UID on success.  */
int vsd_alloc_uid (struct vsd_vs_map *map, int uid_size)
{
  int start_uid, x, found, range_size;
  char *uid_range;

  if (uid_size < 0)
    {
      errno = EINVAL;
      return -1;
    }

  /* No server map so assume we are the first.  */
  if (map == NULL)
    return MIN_UID;

  range_size = ((MAX_UID - MIN_UID) / CHUNK);
  uid_range = (char *) alloca (range_size + 1);
  if (uid_range == NULL)
    return -1;

  memset (uid_range, 0, range_size);

  /* Round up uid_size to the next chunk size.  */
  uid_size = ((uid_size + CHUNK) / CHUNK);

  /* Create an map of allocated uid chunks.  */
  for (x = 0; x < map->servers; x++)
    memset (uid_range + ((map->server[x].startuid - MIN_UID) / CHUNK),
	    1, map->server[x].users / CHUNK);

  found = 0;
  start_uid = 0;
  while (! found && start_uid <= range_size)
    {
      /* Search for an unallocated chunk.  */
      if (uid_range[start_uid] == 0)
	{
	  /* Found one. Is it large enough ?  */
	  found = 1;
	  for (x = 0; x <= uid_size - 1; x++)
	    if (uid_range[start_uid + x] == 1)
	      {
		/* Free chunk is not large enough.  */
		found = 0;
		break;
	      }
	}

      start_uid ++;
    }

  return (! found) ? -1 : ((start_uid - 1) * CHUNK) + MIN_UID;
}

