/* Wrapper program to restrict access to a binary to users who
   have the necessary privilege.
   Copyright (c) 1999 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <pwd.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/resource.h>

#include "libvsd.h"

/* Global variables. These are defined on-the-fly by the program that
   customises the restrict wrapper for the specified binary.   */

/* Name of the vsd privilege that the user must be a member of.
   The vsd privileges and access-control-lists are stored in
   /etc/vsd/priv.  */
extern char *privilege;

/* Length of the main binary in bytes.  */
extern unsigned long program_length;

/* Pointer to the program data. This will be copied to disk by the
   wrapper program and executed.  */
extern unsigned long *program_data;

int main (int argc, char **argv, char **environ)
{
  struct passwd *pw;
  int fd, pid, status;
  char template[20], fname[128];
  struct rlimit lim;

  /* Limit CPU time to 60 seconds.  */
  lim.rlim_cur = 60;
  lim.rlim_max = 60;
  setrlimit (RLIMIT_CPU, &lim);

  /* Limit memory to 48Mb.  */
  lim.rlim_cur = 48*1024*1024;
  lim.rlim_max = 48*1024*1024;
  setrlimit (RLIMIT_DATA, &lim);

#if 0
  /* Limit number of spawned processes.  */
  lim.rlim_cur = 8;
  lim.rlim_max = 8;
  setrlimit (RLIMIT_NPROC, &lim);
#endif

  /* Set core dump files to 0 bytes.  */
  lim.rlim_cur = 0;
  lim.rlim_max = 0;
  setrlimit (RLIMIT_CORE, &lim);

  pw = getpwuid (getuid ());
  if (pw == NULL)
    {
      fprintf (stderr, "Permission denied - user unknown\n");
      return 1;
    }

  if (vsd_priv_access (pw->pw_name, privilege))
    {
      fprintf (stderr, "Permission denied (user doesn't have privilege %s)\n",
	       privilege);
      return 1;
    }

  strcpy (template, "/tmp/wrpXXXXXX");
  mktemp (template);
  /* Open the file write-only.  */
  fd = open (template, O_CREAT | O_WRONLY, 0700);
  if (fd == -1)
    {
      fprintf (stderr, "opening %s failed: %s\n", template, strerror (errno));
      return 1;
    }

  if (write (fd, &program_data, program_length) != program_length)
    {
      int save_errno = errno;
      close (fd);

      if (save_errno == EDQUOT)
	{
	  fprintf (stderr, "\ndisk quota exceeded - you need %dK of disk space to execute this program\n",
		   program_length / 1024);
	  unlink (template);
	  return 1;
	}

      fprintf (stderr,
	       "\nCouldn't write %s. You may have exceeded your quota allocation.\nYou need a minimum of %dK free storage space to execute this program.\n",
	       template, program_length / 1024);
      unlink (template);
     return 1;
    }
  close (fd);

  fd = open (template, O_RDONLY);
  status = 0;
  pid = fork ();
  if (pid == 0)
    {
      /* Child side.  */
      execve (template, argv, environ);
      printf ("can't execute %s: %s\n", template, strerror (errno));
      _exit (127);
    }
  else
    {
      /* Parent side.  */
      struct timespec period, remain;

      period.tv_sec = 0;
      period.tv_nsec = 350000000;
      nanosleep (&period, &remain);

      unlink (template);
      waitpid (pid, &status, 0);
    }

  return status;
}
 
