/* Authentication for the virtual server daemon.
   Copyright (c) 1999 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <netdb.h>
#include <unistd.h>
#ifdef HAVE_LIBPAM
#include <security/pam_appl.h>
#endif
#include <sys/socket.h>
#include <sys/syslog.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include "vsd.h"

#ifdef HAVE_LIBPAM
struct checkpw_cred
{
  const char *uname;
  const char *pass;
};
#endif

#ifdef WITH_OPENSSL
#include "rsa.h"
#include "crypto.h"
#include "x509.h"
#include "pem.h"
#include "ssl.h"
#include "err.h"

SSL_CTX *ssl_ctx;
SSL *ssl_ssl;

#define SSL_KEYS "/etc/vsd/certs"

int authenticate_server (struct connection *vc)
{
  int err;
  X509 *client_cert;
  SSL_METHOD *ssl_method;

  SSL_load_error_strings ();
  ssl_method = SSLv3_server_method ();

  ssl_ctx = SSL_CTX_new (ssl_method);
  if (ssl_ctx == NULL)
    return -1;

  err = SSL_CTX_use_RSAPrivateKey_file (ssl_ctx,
					SSL_KEYS "/plain-key.pem",
					SSL_FILETYPE_PEM);
  if (err == -1)
    {
      ERR_print_errors_fp (stderr);
      return -1;
    }

  err = SSL_CTX_use_certificate_file (ssl_ctx,
				      SSL_KEYS "/plain-cert.pem",
				      SSL_FILETYPE_PEM);
  if (err == -1)
    {
      ERR_print_errors_fp (stderr);
      return -1;
    }

  ssl_ssl = SSL_new (ssl_ctx);
  if (ssl_ssl == NULL)
    return -1;
  SSL_set_fd (ssl_ssl, fileno (stdin));
  err = SSL_accept (ssl_ssl);
  if (err == -1)
    {
      ERR_print_errors_fp (stderr);
      return -1;
    }

  /* Get client's certificate.  */
  client_cert = SSL_get_peer_certificate (ssl_ssl);
  if (client_cert != NULL)
    {
      char subject[256], issuer[256];

      X509_NAME_oneline (X509_get_subject_name (client_cert),
			 subject, sizeof (subject));
      X509_NAME_oneline (X509_get_issuer_name (client_cert),
			 issuer, sizeof (issuer));

      syslog (LOG_INFO, "Client certificate: subj=%s, issuer=%s",
	      subject, issuer);
      Free (subject);
      Free (issuer);
      X509_free (client_cert);
    }
  else
    syslog (LOG_INFO, "Client does not have certificate");

  return 1;
}

#else /* ! WITH_OPENSSL */

int authenticate_server (struct connection *vc)
{
  struct sockaddr_in sin;
  int sin_len = sizeof (struct sockaddr_in);
  int status;
  char *client;
  char connection_originate[64];
#ifndef ENABLE_DEBUG
  char tmp[256];
  struct hostent *hp;
#endif

  status = getpeername (0, (struct sockaddr *) &sin, (void *) &sin_len);
  if (status == 0)
    client = inet_ntoa (sin.sin_addr);
#ifndef ENABLE_DEBUG
  else
    return 0;
#endif

  strcpy (connection_originate, client);

#if 0
  syslog (LOG_ERR, "connect from %s\n", connection_originate);
#endif

#ifdef ENABLE_DEBUG
  vc->login_timeout = 60;
  vc->command_timeout = 60;
#else
  /* Was connection from support.dsvr.co.uk ? */
  if (strcmp (connection_originate, "195.89.164.211") == 0)
    {
      /* If connecting from the dsvr support site, set 10 second timeouts.  */
      vc->login_timeout = 10;
      vc->command_timeout = 10;
    }
  else
    {
      if (gethostname (tmp, sizeof (tmp)) == -1)
	return 0;

      hp = gethostbyname (tmp);
      if (hp == NULL)
	return 0;

      client = inet_ntoa (**((struct in_addr **) hp->h_addr_list));

      if (strcmp (connection_originate, client) != 0)
	{
	  syslog (LOG_ALERT, "access denied from %s\n", connection_originate);
	  return 0;
	}
    }
  vc->login_timeout = 4;
  vc->command_timeout = 1;
#endif
  return 1;
}
#endif /* WITH_OPENSSL */

/* Parse PASS command
 * Accepts: pointer to command argument
 * Returns: new state
 */

int login (struct connection *vc, const char *t)
{
  /* flush old passowrd */
#if 0
  if (password)
    free (password);
#endif

  if (t == NULL || *t == '\0')
    {
      /* No password given.  */
      output (vc, "-ERR Missing password argument\n");
      return STATE_AUTHORISATION;
    }

  /* Copy password.  */
  if (server_login (vc, t))
    {
      syslog (LOG_INFO, "Login to vs=%.80s user=%.80s host=%.80s",
	      vc->virtual_server, vc->user_name,
	      vc->client_host);
      return STATE_TRANSACTION;
    }

  /* Login failed.  */
  return STATE_AUTHORISATION;
}

int server_login (struct connection *vc, const char *password)
{
#if 0
  struct passwd *pw;
  char vs_user[128];
#endif

  if (strlen (vc->user_name) >= 80)
    syslog (LOG_ALERT | LOG_AUTH,
	    "System break-in attempt to vs=%.80s from host=%.80s",
	    vc->virtual_server, vc->client_host);

  if (strlen (vc->user_name) >= 40)
    syslog (LOG_ALERT | LOG_AUTH,
	    "System break-in attempt to vs=%.80s from host=%.80s",
	    vc->virtual_server, vc->client_host);

  return 1;

#if 0
  if (strcmp (vc->virtual_server, "localhost") == 0)
    strcpy (vs_user, vc->user_name);
  else
    {
      /* Now first of all the only person allowed to log into this system
	 is the 'admin' user of a virtual server.  */
      if (strcmp (vc->user_name, "admin") != 0)
	{
	  syslog (LOG_WARNING | LOG_AUTH,
		  "Login attempt to vs=%.80s user=%.80s from host=%.80s failed",
		  vc->virtual_server, vc->user_name, vc->client_host);
	  return 0;
	}
    }

  pw = getpwnam (vs_user);
  if (pw != NULL)
    {
#if 0
      printf ("user = '%s', password = '%s'\n", vs_user, password);
#endif
      pw = checkpw (pw, password);
      if (pw != NULL)
	return 1;

      if (*password == ' ')
	{
	  pw = checkpw (pw, password + 1);
	  if (pw != NULL)
	    return 1;
	}
    }

  syslog (LOG_INFO, "Login failure vs=%.80s user=%.80s host=%.80s",
	  vc->virtual_server, vc->user_name, vc->client_host);
  sleep (3);
  return 0;
#endif
}

#ifdef HAVE_LIBPAM
int checkpw_conv (int num_msg, const struct pam_message **msg,
		  struct pam_response **resp, void *appdata_ptr)
{
  int i;
  struct checkpw_cred *cred;
  struct pam_response *reply;

  cred = (struct checkpw_cred *) appdata_ptr;
  reply = malloc (sizeof (struct pam_response) * num_msg);
  for (i = 0; i < num_msg; i++)
    switch (msg[i]->msg_style)
      {
      case PAM_PROMPT_ECHO_ON:
	/* Assume want user name.  */
	reply[i].resp_retcode = PAM_SUCCESS;
	reply[i].resp = strdup (cred->uname);
	break;
      case PAM_PROMPT_ECHO_OFF:
	/* Assume want password.  */
	reply[i].resp_retcode = PAM_SUCCESS;
	reply[i].resp = strdup (cred->pass);
	break;
      case PAM_TEXT_INFO:
      case PAM_ERROR_MSG:
	fputs (msg[i]->msg, stdout);
	reply[i].resp_retcode = PAM_SUCCESS;
	reply[i].resp = NULL;
	break;
      default:
	free (reply);
	return PAM_CONV_ERR;
      }

  *resp = reply;
  return PAM_SUCCESS;
}
#endif

struct passwd *checkpw (struct passwd *pw, const char *pass)
{
#ifdef HAVE_LIBPAM
  pam_handle_t *hdl;
  struct pam_conv conv;
  struct checkpw_cred cred;

  conv.conv = &checkpw_conv;
  conv.appdata_ptr = &cred;
  cred.uname = pw->pw_name;
  cred.pass = (char *) pass;

  if (pam_start ("vsd", pw->pw_name, &conv, &hdl) == PAM_SUCCESS
      && pam_authenticate (hdl, 0) == PAM_SUCCESS
      && pam_acct_mgmt (hdl, 0) == PAM_SUCCESS
      && pam_setcred (hdl, PAM_ESTABLISH_CRED) == PAM_SUCCESS)
    {
      pam_end (hdl, PAM_SUCCESS);
      return pw;
    }

  /* Access denied.  */
  pam_end (hdl, PAM_AUTH_ERR);
  return NULL;
#endif

  return pw;
}
