/* Input/output handling.
   Copyright (c) 1999 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* Protocol input format:
     virtual-server {
       command [args]
       command {
     <data>
       }
     }
     virtual-server {
       command [args]
       command [args]
     }
     EOF

   Output format:
     virtual-server {
       command {
         result-code/result-data
       }
       command {
         result-code/result-data
       }
     }
     EOF

   Please see freevsd/doc/freevsd-protocol.txt for a fuller description
   of the protocol.  */

#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "vsd.h"
#include "libvsd.h"

/* Get the next line of input from the connected socket.  */
static int input (struct connection *vc, int state)
{
  int pos, ret;
  char *p;

  alarm (vc->command_timeout);

  if (vc->input_buffer == NULL)
    {
      vc->input_bufsiz = 256;
      vc->input_buffer = (char *) malloc (vc->input_bufsiz);
    }

  pos = 0;
  while (1)
    {
      if (vc->input_bufsiz <= pos)
	{
	  /* Need more space.  */
	  vc->input_bufsiz += 256;
	  vc->input_buffer = (char *) realloc (vc->input_buffer,
					       vc->input_bufsiz);
	}

      /* There is a buffering problem that I can't seem to remember how
	 to get around. So instead, read the stream one byte at a time.  */
      while ((ret = read (fileno (stdin), vc->input_buffer + pos, 1)) == -1
	     && errno == EINTR)
	;

      /* Replace this with a select call.  */
      if (ret == 0)
	{
	  /* Sleep for a short period to prevent a CPU resource drain.  */
	  struct timespec period, remain;
	  period.tv_sec = 0;
	  period.tv_nsec = 200000000;
	  nanosleep (&period, &remain);
	}

      /* cp = strstr (vc->input_buffer, "\n"); */
      /* if (cp) */

      pos += ret;
      if (vc->input_buffer[pos - ret] == '\n')
	{
	  if (vc->input_buffer[pos - ret - 1] == '\r')
	    pos --;
	  vc->input_buffer[pos - ret] = '\0';
	  /* *cp = '\0'; */
	  alarm (0);
	  return 0;
	}
    }
  alarm (0);

  return 1;
}

/* Allocate memory of size `bytes' bytes.  If ptr is not NULL, then assume
   we are re-allocating memory.  */
static void *xalloc (void *ptr, size_t bytes)
{
  if (! ptr)
    return calloc (bytes, sizeof (char));
  return realloc (ptr, bytes);
}

/* Parse incoming data, storing it in an `io_requests' structure.  */
struct io_requests *io_parse_input (struct connection *vc)
{
  struct io_requests *req = NULL;
  int level, req_num = 0, cmd_num;

  /* Nesting level. At level 0 we expect a virtual-server name. Increment/
     decrement when we come across a `{'/`}' respectively. At level 1
     we expect commands.  */
  level = 0;

  /* Process data one line at a time, the terminator is the string "EOF".  */
  while (! input (vc, STATE_TRANSACTION))
    {
      char *in = vc->input_buffer;
      char **vec = NULL;
      size_t vecc = 0;

      /* Nesting level 2 is where we accept raw data. So it isn't
	 a good idea to split the line up into vectors.  */
      if (level != 2)
	{
	  /* Skip blanks.  */
	  while (*in && isspace (*in))
	    in++;
	  if (!*in)
	    continue; /* Blank line.  */
	  /* EOF terminates an input stream.  */
	  if (strcmp (in, "EOF") == 0)
	    break;

	  /* Split space-seperated arguments into an argument vector.  */
	  vec = vsd_argv_parse (' ', in, &vecc);
	  if (vec == NULL)
	    continue;
	}

      if (level == 0)
	{
	  /* Expect `virtual-server {'.  */
	  if (! strcmp (vec[0], "<root>")
	      || vsd_getmapent (vc->map, vec[0], NULL))
	    {
	      req = (struct io_requests *)
		xalloc (req, (req_num + 1) * sizeof (struct io_requests));
	      memset (&req[req_num], 0, sizeof (struct io_requests));
	      req[req_num].vs = strdup (vec[0]);
	      cmd_num = 0;
	      level ++;
	    }
	  /* vsd_argv_parse allocates memory for the argument vector.  */
	  free (vec);
	}
      else if (level == 1)
	{
	  /* Expect either a command or a '}'.  */
	  if (*in == '}')
	    {
	      /* And so ends this block. Increment request number.  */
	      req[req_num++].actionc = cmd_num;
	      level --;
	    }
	  else
	    {
	      int x;

	      /* Found a command. Create space for it and make a copy of
		 the command and its arguments.  */
	      req[req_num].actionv = (struct req_command *)
		xalloc (req[req_num].actionv,
			(cmd_num + 1) * sizeof (struct req_command));

	      req[req_num].actionv[cmd_num].argv = (char **)
		malloc (sizeof (char *) * (vecc + 1));
	      if (*vec[vecc - 1] == '{')
		{
		  /* Command has extra data.  */
		  level ++;
		  /* Avoid adding curly bracket to the command line.  */
		  vecc --;
		}
	      for (x = 0; x < vecc; x++)
		req[req_num].actionv[cmd_num].argv[x] = strdup (vec[x]);

	      /* req[req_num].actionv[cmd_num].argv = vec; */
	      req[req_num].actionv[cmd_num].argc = vecc; /* Arg count.  */
	      /* Don't change command numbers if we have just gone up a
		 level as we want to associate the extra data with this
		 command.  */
	      if (level == 1)
		cmd_num ++; 
	    }
	  /* vsd_argv_parse allocates memory for the argument vector.  */
	  free (vec);
	}
      else if (level == 2)
	{
	  if (*in == '}')
	    {
	      /* End of data segment.  */
	      req[req_num].actionv[cmd_num].data = strdup (vc->io_buffer);
	      io_buffer_free (vc);
	      level --;
	      cmd_num ++;
	    }
	  else
	    io_buffer_store (vc, "%s\n", in);
	}
    }
  
  vc->requests = req;
  vc->requests_count = req_num;
  return req;
}

/* Output collected VSD results in a suitable format for the client.  */
void io_generate_output (struct connection *vc)
{
  struct io_results *res = vc->results;
  int i, j;

  switch (vc->io_output_format)
    {
    case 1:
      /* Output format:
	 <virtual server> <cmd name> <result data>
	 <virtual server> <cmd name> <result data>
      */
      for (i = 0; i < vc->results_count; i++)
	for (j = 0; j < res[i].cmdc; j++)
	  output (vc, "%s %s %s\n",
		  res[i].vs, res[i].cmdv[j].cmd, res[i].cmdv[j].data);
      break;
    default:
      for (i = 0; i < vc->results_count; i++)
	{
	  output (vc, "%s {\n", res[i].vs);
	  for (j = 0; j < res[i].cmdc; j++)
	    {
	      output (vc, "  %s {\n", res[i].cmdv[j].cmd);
	      output (vc, "%s\n",
		      (res[i].cmdv[j].data) ? res[i].cmdv[j].data : "");
	      output (vc, "  }\n");
	    }
	  output (vc, "}\n");
	}
      break;
    }

  output (vc, "EOF\n");
}

/* Store the data collected by `io_buffer_store' into the results
   structure.  */
void io_collect_output (struct connection *vc, const char *command)
{
  struct io_results *res = vc->results;
  int i;

  /* Try and find a matching virtual server in the results collected
     so far. If we do, then add the command and its data to the `command'
     struct.  */
  for (i = 0; i < vc->results_count; i++)
    if (strcmp (vc->virtual_server, res[i].vs) == 0)
      break;

  if (i == vc->results_count)
    {
      /* No match found.  Create a new slot.  */
      vc->results_count ++;
      res = (struct io_results *) xalloc (res, vc->results_count
					  * sizeof (struct io_results));
      memset (&res[i], 0, sizeof (struct io_results));
      res[i].vs = strdup (vc->virtual_server);
    }
  /* Found a matching VS. Add a new command.  */
  res[i].cmdv = (struct res_command *) xalloc (res[i].cmdv, (res[i].cmdc + 1)
					   * sizeof (struct res_command));
  res[i].cmdv[res[i].cmdc].cmd = strdup (command);
  res[i].cmdv[res[i].cmdc].data = ((vc->io_buffer)
				   ? strdup (vc->io_buffer)
				   : NULL);
  res[i].cmdc ++;
  vc->results = res;
  io_buffer_free (vc);
}

/* Collect data specified by the format string `fmt' and its arguments
   into a extendable buffer.  */
void io_buffer_store (struct connection *vc, const char *fmt, ...)
{
  va_list ap;
  int nchars, freespace;

  va_start (ap, fmt);
  if (! vc->io_buffer_size)
    {
      /* Size of the buffer.  */
      vc->io_buffer_size = 1024;
      /* Pointer to the start of the buffer.  */
      vc->io_buffer = (char *) malloc (vc->io_buffer_size);
      /* Pointer to the end of the last bit of text written out.  */
      vc->io_bufferp = vc->io_buffer;
    }
  /* Calculate actual space available in buffer for writing a string.  */
  freespace = vc->io_buffer_size - (vc->io_bufferp - vc->io_buffer);
  nchars = vsnprintf (vc->io_bufferp, freespace, fmt, ap);
  if (nchars >= freespace)
    {
      int i = vc->io_bufferp - vc->io_buffer;

      /* Not enough free space. Allocate some more.  */
      vc->io_buffer_size += nchars + 2 - freespace;
      vc->io_buffer = (char *) realloc (vc->io_buffer, vc->io_buffer_size);
      vc->io_bufferp = vc->io_buffer + i;
    }
  /* Finally store the string.  */
  vc->io_bufferp += vsnprintf (vc->io_bufferp, vc->io_buffer_size, fmt, ap);
  va_end (ap);
}

/* Free the buffer used by `io_buffer_store'.  */
void io_buffer_free (struct connection *vc)
{
  if (vc->io_buffer_size)
    {
      free (vc->io_buffer);
      vc->io_buffer = vc->io_bufferp = NULL;
      vc->io_buffer_size = 0;
    }
}

#if 0
struct io_requests *io_parse_output (struct connection *vc)
{
  struct io_results *res = NULL;
  int level, res_num, cmd_num;

  level = 0;
  while (! input (vc, STATE_TRANSACTION))
    {
      char *in = vc->input_buffer;
      char **vec = NULL;
      size_t vecc = 0;

      if (level != 2)
	{
	  /* Skip blanks.  */
	  while (*in && isspace (*in))
	    in++;
	  if (! *in)
	    continue;
	  if (strcmp (in, "EOF") == 0)
	    break;

	  /* Split arguments into an argument vector.  */
	  vec = vsd_argv_parse (' ', in, &vecc);
	  if (vec == NULL)
	    continue;
	}

      if (level == 0)
	{
	  /* Expect `virtual server {'.  */
	  if (! vsd_getmapent (vc->map, vec[0], NULL))
	    { /* Virtual server doesn't exist.  */ }
	  else
	    {
	      res = (struct io_results *)
		xalloc (res, (res_num + 1) * sizeof (struct io_results));
	      memset (&res[res_num], 0, sizeof (struct io_results));
	      res[res_num].vs = strdup (vec[0]);
	      cmd_num = 0;
	      level ++;
	    }
	  free (vec);
	}
      else if (level == 1)
	{
	  /* Expect a command or a `}'.  */
	  if (*in == '}')
	    {
	      res[res_num++].cmdc = cmd_num;
	      level --;
	    }
	  else
	    {
	      int x;

	      res[res_num].cmdv = (struct res_command *)
		xalloc (res[res_num].cmdv,
			(res_num + 1) * sizeof (struct res_command));
	      res[res_num].cmdv[cmd_num].cmd = strdup (vec[0]);
	      if (*vec[vecc - 1] == '{')
		{
		  level ++;
		  vecc --;
		}
	      if (level == 1)
		cmd_num++;
	    }
	}
      else if (level == 2)
	{
	  if (*in == '}')
	    {
	      /* End of data segment.  */
	      res[res_num].cmdv[cmd_num].data = strdup (vc->io_buffer);
	      io_buffer_free (vc);
	      level --;
	      cmd_num ++;
	    }
	  io_buffer_store (vc, "%s\n", in);
	}
    }
}
#endif



