/* Module command interface (client interface)
   Copyright (c) 1999 Designer Servers Ltd.
   Written by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD).

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "module.h"
#include "mod_quota.h"
#include "vsd_lib.h"
#include "libvsd.h"
#include "vsd.h"

struct command_tag mod_quota_command_list[] = 
{
  {
    "quota_vsset", CMD_QUOTA_VSSET, ADMIN_HOST,
    "Set disk quota limits for a virtual server",
    3, "quota_vsset <host server> <virtual server> [<+|->]<bytes>[<K|M>]",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "quota_vsget", CMD_QUOTA_VSGET, ADMIN_HOST,
    "Get disk quota limits for a virtual server",
    2, "quota_vsget <host server> <virtual server>",
    NULL, 0, NULL /* for vsclients  */
  },
  {
    "quota_stats", CMD_QUOTA_STATS, ADMIN_VS | ADMIN_CLIENT,
    "Obtain quota stats for all users",
    2, "quota_stats <host server> <virtual server>",
    "quotastats", 0, "quotastats [<username>]" /* for vsclients */
  },
  {
    "quota_userset", CMD_QUOTA_USERSET, ADMIN_VS | ADMIN_CLIENT,
    "Set quota limits for a user",
    4, "quota_userset <host server> <virtual server> <user> [<+|->]<bytes>[<K|M>]",
    "setquota", 2, "setquota <user> [<+|->]<bytes>[<K|M>]" /* for vsclients */
  },
  {
    "quota_userget", CMD_QUOTA_USERGET, ADMIN_VS,
    "Get quota limits for a user",
    3, "quota_userget <host server> <virtual server> <user>",
    NULL, 0, NULL /* for vsclients */
  },
  {
    NULL, 0, 0,
    NULL, 0, NULL,
    NULL, 0, NULL
  }
};

/* If the command is not implemented by this module or at the level `level'
   then return `notexist'. If there are an invalid number of arguments
   then return the module syntax. If the syntax is correct then
   return NULL.  */
const char *mod_quota_check_syntax (int level, const char *command,
				    int argc, char *argv[])
{
  struct command_tag *list;

  list = module_lookup_command (mod_quota_command_list, level, command);

  /* Command doesn't exist.  */
  if (! list)
    return "notexist";

  if (level == ADMIN_CLIENT)
    {
      if (argc < list->client_command_args)
	return list->client_command_syntax;

      return NULL;
    }

  if (argc != list->num_args)
    return list->syntax;

  return NULL;
}

int mod_quota_list_help (int level, const char *command)
{
  struct command_tag *list;

  if (command)
    {
      list = module_lookup_command (mod_quota_command_list, level, command);
      if (list == NULL)
	return 1;

      fprintf (stderr, "  %14s - %s\n", list->tag, list->help);
    }
  else
    {
      int y = 0;

      list = mod_quota_command_list;
      if (list == NULL)
	return 1;

      while (list[y].tag)
	{
	  if (list[y].level & level)
	    fprintf (stderr, "  %14s - %s\n", list[y].tag, list[y].help);
	  y++;
	}
    }

  return 0;
}

char *mod_quota_do_cmd (vsd_server *server, int level,
			const char *cmd, char **argv)
{
  struct command_tag *list;
  char *status, *line;
  int uid, hardb, softb, curb, data;
  dstring_t str;

  list = module_lookup_command (mod_quota_command_list, level, cmd);
  if (! list)
    return "notexist";

  switch (list->token)
    {
    case CMD_QUOTA_VSSET:
      vsd_send_command (server, NULL, "QUOTA_VSSET %s %s", argv[1], argv[2]);
      status = vsd_process_reply (server);
      break;
    case CMD_QUOTA_VSGET:
      vsd_send_command (server, NULL, "QUOTA_VSGET %s", argv[1]);
      status = vsd_process_reply (server);
      break;
    case CMD_QUOTA_STATS:
      str = dstring_new (512);
      vsd_send_command (server, NULL, "QUOTA_STATS");

      if (vsd_interface () == CLI_INTERFACE)
	{
	  dstring_append (str, "=================+========+=============+============\n");
	  dstring_append (str, "User name        | Uid    | Limit       | Current use\n");
	  dstring_append (str, "=================+========+=============+============\n");
	}

      status = vsd_process_reply (server);
      while ((line = io_getline ('\n', &status)) != NULL)
	{
	  if (strstr (line, "user=") != NULL)
	    {
	      if (vsd_interface () == WEB_INTERFACE)
		dstring_fmt_append (str, "%s\n", line);
	      else
		{
		  char user_name[128];

		  sscanf (line,
			  "user=%[^ ] uid=%d hardb=%d softb=%d curb=%d\n",
			  user_name, &uid, &hardb, &softb, &curb);

		  if (hardb == 0)
		    dstring_fmt_append (str, "%17s|%8d|%13s|%11dK\n",
					user_name, uid, "unlimited", curb);
		  else
		    dstring_fmt_append (str, "%17s|%8d|%12dK|%11dK\n",
					user_name, uid, softb, curb);
		}
	    }
	  else
	    {
	      sscanf (line, "hardb=%d softb=%d curb=%d\n",
		      &hardb, &softb, &curb);
	      if (vsd_interface () == CLI_INTERFACE)
		{
		  dstring_append (str, "=================+========+=============+============\n");
		  if (softb == 0)
		    dstring_fmt_append (str, "%17s|        |%13s|%11dK\n",
					"Allocated/used", "unlimited", curb);
		  else
		    dstring_fmt_append (str, "%17s|        |%12dK|%11dK\n",
					"Allocated/used", softb, curb);
		}
	      else
		dstring_fmt_append (str, "subtotal: softb=%d, curb=%d\n",
				    softb, curb);
	    }
	  free (line);
	}

#if 0
      vsd_send_command (server, NULL, "QUOTA_VSGET");
      status = vsd_process_reply (server);
      sscanf (status, "vsquota=%d", &curb);
      if (vsd_interface () == CLI_INTERFACE)
	{
	  if (curb == 0)
	    dstring_fmt_append (str, "%17s|        |%13s|\n",
				"Server limit", "unlimited");
	  else
	    dstring_fmt_append (str, "%17s|        |%12dK|\n",
				"Server limit", curb);
	  dstring_append (str, "=================+========+=============+============\n");
	}
      else
	dstring_fmt_append (str, "limit: curb=%d\n", curb);
#endif

      status = dstring_yield (str);
      dstring_free (str);
      break;
    case CMD_QUOTA_USERSET:
      vsd_send_command (server, NULL, "QUOTA_USERSET %s %s", argv[2], argv[3]);
      status = vsd_process_reply (server);
      break;
    case CMD_QUOTA_USERGET:
      str = dstring_new (512);
      
      vsd_send_command (server, NULL, "QUOTA_USERGET %s", argv[2]);
      status = vsd_process_reply (server);
      if (status)
	return status;

      sscanf (status, "uid=%d, hardb=%d, softb=%d, curb=%d",
	      &uid, &hardb, &softb, &curb);
      if (vsd_interface () == CLI_INTERFACE)
	{
	  if (hardb == 0)
	    dstring_fmt_append (str,
				"No disk quota is set for user %s (uid %d)\n",
				argv[2], uid);
	  else
	    {
	      dstring_fmt_append (str, "Disk quota for user %s (uid %d)\n",
				  argv[2], uid);
	      dstring_fmt_append (str, "        Limit : %dK\n", softb);
	      dstring_fmt_append (str, "  Current use : %dK\n", curb);
	    }
	}
      else
	dstring_fmt_append (str, "%s", status);

      status = dstring_yield (str);
      dstring_free (str);
      break;
    }

  return status;
}

module quota_module =
{
  mod_quota_check_syntax,
  mod_quota_list_help,
  mod_quota_do_cmd
};
