/* Quota command implementation.
   Copyright (c) 1999 Designer Servers Ltd.
   Written by Nick Burrett <nick@dsvr.co.uk>

   This file is part of the Virtual Server Administrator (FreeVSD).

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/syslog.h>

#ifdef HAVE_SYS_QUOTA_H
#include <sys/quota.h>
#else
#ifdef HAVE_UFS_UFS_QUOTA_H
#include <ufs/ufs/quota.h>
#endif
#endif

#include "vsd.h"
#include "libvsd.h"
#include "quota.h"

/* DEV is passed to quotactl.  */
#if defined(HAVE_SYS_QUOTA_H)
#define DEV(map,vs) (map)->partition[(vs)->partition].device
#elif defined(HAVE_UFS_UFS_QUOTA_H)
#define DEV(map,vs) (map)->partition[(vs)->partition].pathname
#endif

static int diskquota (int cmd, const char *special, int uid, caddr_t addr)
{
#ifdef HAVE_SYS_QUOTA_H
  /* Linux style.  */
  return quotactl (cmd, special, uid, addr);
#else
#ifdef HAVE_UFS_UFS_QUOTA_H
  /* BSD style.  */
  return quotactl (special, cmd, uid, addr);
#endif
#endif
}

/* Get quota stats for all users on the virtual server.
   If uid is NULL then don't create a uid list.
   If puser is NULL then don't create user quota stats.
   If total is NULL then don't tally totals.  */
static int vs_user_stats (struct connection *vc,
			  int **puid, void **puser_, void *total_,
			  const char *server_name, const char *device)
{
  FILE *stream;
  struct passwd *pwd;
  struct dqblk user_quota;
  int count, max, status;
  int *uid;
  struct dqblk *user;
  struct dqblk **puser = (struct dqblk **)puser_;
  struct dqblk *total = (struct dqblk *)total_;
  uid_t min_uid, max_uid;

  if (puid == NULL && puser == NULL && total == NULL)
    return 0;

  stream = vsd_openpw (vc, server_name);
  vsd_map_uid_range (vc->map, server_name, &min_uid, &max_uid);

  if (stream == NULL)
    return 0;

  max = 16;
  count = 0;

  uid = (int *) malloc (max * sizeof (int));
  user = (struct dqblk *) malloc (max * sizeof (struct dqblk));

  if (total != NULL)
    {
      total->dqb_bhardlimit = 0;
      total->dqb_bsoftlimit = 0;
      total->dqb_curblocks = 0;
    }

  /* Look at each /etc/passwd entry but only find quota information
     for users on the virtual server.  */
  while ((pwd = fgetpwent (stream)))
    if (pwd->pw_uid >= min_uid && pwd->pw_uid <= max_uid)
      {
	if (puid != NULL)
	  uid[count] = pwd->pw_uid;

	status = diskquota (QCMD (Q_GETQUOTA, USRQUOTA), device,
			    pwd->pw_uid, (caddr_t) &user_quota);
	if (status == 0)
	  {
	    if (puser != NULL)
	      user[count] = user_quota;

	    if (count >= max - 1)
	      {
		max += 16;
		if (puid != NULL)
		  uid = (int *) realloc (uid, max * sizeof (int));
		if (puser != NULL)
		  user = (struct dqblk *) realloc (user,
						   max * sizeof (struct dqblk));
	      }

	    if (total != NULL)
	      {
		total->dqb_bhardlimit += user_quota.dqb_bhardlimit;
		total->dqb_bsoftlimit += user_quota.dqb_bsoftlimit;
		total->dqb_curblocks += user_quota.dqb_curblocks;
	      }
	  }
	else
	  {
	    if (puser != NULL)
	       memset (&user[count], 0, sizeof (struct dqblk));
	  }

	count ++;
      }

  vsd_closepw (stream);

  if (puid)
    *puid = uid;
  if (puser)
    *puser = user;

  return count;
}

/* Return quota limits for a user.  */
void quota_userget (struct connection *vc, int argc, char *argv[])
{
  struct dqblk quota;
  int status;
  uid_t user_uid;

  if (is_name_valid (vc, USER, NULL, argv[0], NAME_DONT_CHECK_LEN))
    return;

  /* Get the user id.  */
  if (user_exists (vc, vc->virtual_server, argv[0], &user_uid))
    {
      io_buffer_store (vc, "-ERR User %s does not exist", argv[0]);
      return;
    }

  /* should be a /dev */
  status = diskquota (QCMD (Q_GETQUOTA, USRQUOTA),
		      DEV (vc->map, vc->vs), user_uid, (caddr_t) &quota);

  if (status == -1)
    {
      syslog (LOG_ERR, "Get quota stats failed: %m");
      io_buffer_store (vc, "-ERR Could not get quota stats");
      return;
    }
  
  io_buffer_store (vc, "uid=%d hardb=%d softb=%d curb=%d", user_uid,
		   quota.dqb_bhardlimit, quota.dqb_bsoftlimit,
		   quota.dqb_curblocks);
}

/* Set quota limits for a user.  */
void quota_userset (struct connection *vc, int argc, char *argv[])
{
  struct dqblk user_quota, total_quota;
  int status, bytes;
  const char *user, *size, *p;
  uid_t user_uid;

  if (is_name_valid (vc, USER, NULL, argv[0], NAME_DONT_CHECK_LEN))
    return;

  user = argv[0];
  size = argv[1];

  /* Get the uid for the user that quotas are to be set for.  */
  if (user_exists (vc, vc->virtual_server, user, &user_uid))
    {
      io_buffer_store (vc, "-ERR User %s does not exist", user);
      return;
    }

  /* Get quota stats for the virtual server.  We require this to ensure
     that an individual quota limit is not set greater than the quota
     limit of the virtual server.  */

  /* Get quota stats for all users on the virtual server.  */
  vs_user_stats (vc, NULL, NULL, &total_quota, vc->virtual_server,
		 DEV (vc->map, vc->vs));

  /* Get quota stats for the user.  */
  status = diskquota (QCMD (Q_GETQUOTA, USRQUOTA), DEV (vc->map, vc->vs),
		      user_uid, (caddr_t) &user_quota);
  if (status == -1 && errno == ESRCH)
    memset (&user_quota, 0, sizeof (user_quota));
  else if (status == -1 && errno != ESRCH)
    {
      syslog (LOG_ERR, "set_quota: get user quota stats failed: %m");
      io_buffer_store (vc, "-ERR Could not get user quota stats");
      return;
    }

  /* Subtract the user's quota from the server's total. We can then
     calculate a maximum allocation for the user.  */
  total_quota.dqb_bhardlimit -= user_quota.dqb_bhardlimit;
  total_quota.dqb_bsoftlimit -= user_quota.dqb_bsoftlimit;
  total_quota.dqb_curblocks -= user_quota.dqb_curblocks;

  /* Parse the size argument.  */
  p = size;
  if (*p == '+' || *p == '-')
    p++; /* Skip the +/- sign for now.  */

  bytes = 0;
  while (isdigit (*p))
    bytes = (bytes * 10) + (*p++ - '0');

  /* Handle a kilobyte/megabyte expression.  */
  if (toupper (*p) == 'M')
    bytes *= 1024;

  /* Now handle the +/- sign.  */
  if (size[0] == '+')
    user_quota.dqb_bsoftlimit += bytes;
  else if (size[0] == '-')
    user_quota.dqb_bsoftlimit -= bytes;
  else if (isdigit (size[0]))
    user_quota.dqb_bsoftlimit = bytes;

  /* Make the user's quota hard limit equal to it's soft limit.  */
  user_quota.dqb_bhardlimit = user_quota.dqb_bsoftlimit;

  /* Ensure user quota lies within the boundaries of the virtual server's
     disk quota.  */
  if (user_quota.dqb_bsoftlimit + total_quota.dqb_bsoftlimit
      > vc->vs->quota * 1024)
    {
      /* We could do this but I think it is a better idea to report an
	 error to the user.
	 user_quota.bsoftlimit = server_quota.bsoftlimit;  */
      io_buffer_store (vc, "-ERR Quota would exceed virtual server allocation");
      return;
    }

  if (user_quota.dqb_bsoftlimit < 0)
    {
      /* User's quota has been set far too low.  Flag as an error.  */
      io_buffer_store (vc, "-ERR User would have a negative quota value");
      return;
    }

  if (user_quota.dqb_bsoftlimit == 0)
    {
      /* Don't let a user set the quota to unlimited.  */
      io_buffer_store (vc, "-ERR User quota limit cannot be set to 0K\n");
      return;
    }

  status = diskquota (QCMD (Q_SETQUOTA, USRQUOTA),
		      DEV (vc->map, vc->vs),
		      user_uid, (caddr_t) &user_quota);
  if (status == -1)
    {
      syslog (LOG_ERR, "Set quota stats failed: %m");
      io_buffer_store (vc, "-ERR Could not set quota stats");
      return;
    }

  /* Synchronise the new quota to disk.  */
  if (diskquota (QCMD (Q_SYNC, 0), DEV (vc->map, vc->vs), 0, (caddr_t) 0) != 0)
    /* Don't report this as an error to the user.  */
    syslog (LOG_ERR, "set_quota: sync to disk failed: %m");
}

/* Get quota statistics for a virtual server and all users on it.  */
void quota_stats (struct connection *vc, int argc, char *argv[])
{
  int *uidv, count;
  struct dqblk *userv, totalv;
  struct passwd *pw;
  FILE *stream;

  count = vs_user_stats (vc, &uidv, (void **) &userv, (void **) &totalv,
			 vc->virtual_server, DEV (vc->map, vc->vs));

  stream = vsd_openpw (vc, vc->virtual_server);
  if (stream == NULL)
    return;

  while (--count >= 0)
    {
      pw = vsd_getpwuid (stream, uidv[count]);
      /* Strip the virtual server name off the user name.  */
      io_buffer_store (vc, "user=%s uid=%d hardb=%d softb=%d curb=%d\n",
		       pw->pw_name, uidv[count],
		       userv[count].dqb_bhardlimit,
		       userv[count].dqb_bsoftlimit,
		       userv[count].dqb_curblocks);
    }

  io_buffer_store (vc, "hardb=%d softb=%d curb=%d",
		   totalv.dqb_bhardlimit, totalv.dqb_bsoftlimit,
		   totalv.dqb_curblocks);

  vsd_closepw (stream);
  free (uidv);
  free (userv);
}

/* Return quota limits for a virtual server.  */
void quota_vsget (struct connection *vc, int argc, char *argv[])
{
  struct vsd_vs *vs;
  int quota;

  vs = vsd_getmapent (vc->map, vc->virtual_server, NULL);
  quota = vs->quota * 1024;
  io_buffer_store (vc, "vsquota=%d", quota);
}

/* Set quota limits for a virtual server.  This call is restricted
   to a host server administrator.  */
void quota_vsset (struct connection *vc, int argc, char *argv[])
{
  struct vsd_vs *vs;
  struct dqblk total_quota;
  int quota, bytes;
  char *p, *size = argv[0];

  vs = vsd_getmapent (vc->map, vc->virtual_server, NULL);
  quota = vs->quota * 1024;

  /* Parse the size argument.  */
  p = size;
  if (*p == '+' || *p == '-')
    p++; /* Skip the +/- sign for now.  */

  bytes = 0;
  while (isdigit (*p))
    bytes = (bytes * 10) + (*p++ - '0');

  /* Handle a kilobyte/megabyte expression.  */
  if (toupper (*p) == 'M')
    bytes *= 1024;

  if (size[0] == '+')
    quota += (bytes / 1024);
  else if (size[0] == '-')
    quota -= (bytes / 1024);
  else
    quota = bytes / 1024;

  /* Get total quota allocation for all users on the virtual server.  */
  vs_user_stats (vc, NULL, NULL, &total_quota, vc->virtual_server,
		 DEV (vc->map, vs));
  if (bytes < total_quota.dqb_bsoftlimit)
    {
      /* Server's quota has been set lower than the total allocation
	 for all users on the server.  */
      io_buffer_store (vc, "-ERR Total user quota allocation is larger than server quota");
      return;
    }

  if (bytes < 0)
    /* Server's quota has been set far too low.  */
    io_buffer_store (vc, "-ERR Server quota would have negative value");
  else
    {
      vs->quota = bytes / 1024;
      vsd_map_save (vc->map);
    }
}
