/* User account manipulation.
   Copyright (c) 1999, 2000 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/syslog.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "vsd.h"
#include "libvsd.h"

#define USE_PROFTPD 1

extern char *stpcpy (char *, const char *);

/* List all users on a virtual server and various details about them.
   This command takes no arguments.  */
void user_list (struct connection *vc, int argc, char *argv[])
{
  struct passwd *pw;
  FILE *pw_stream;
  uid_t min_uid, max_uid;

  pw_stream = vsd_openpw (vc, vc->virtual_server);
  if (pw_stream == NULL)
    io_buffer_store (vc, "-ERR Could not open %s password file",
		     vc->virtual_server);
  else
    {
      /* Obtain a list of all users on the virtual server.  */
      vsd_get_uidgid (vc->server_root, &min_uid, &max_uid, NULL, NULL);
      
      while ((pw = fgetpwent (pw_stream)))
	if (pw->pw_uid >= min_uid && pw->pw_uid <= max_uid)
	  io_buffer_store (vc, "name=%s uid=%d home=\"%s\" shell=\"%s\"\n",
			   pw->pw_name, pw->pw_uid, pw->pw_dir,
			   pw->pw_shell);

      vsd_closepw (pw_stream);
    }
}

/* List all user groups on a virtual server and various details about them.
   This command takes no arguments.  */
void user_group_list (struct connection *vc, int argc, char *argv[])
{
  struct group *grp;
  FILE *gr_stream;
  gid_t min_uid, max_uid;

  gr_stream = vsd_opengrp (vc->server_root);
  if (gr_stream == NULL)
    io_buffer_store (vc, "-ERR Could not open %s group file",
		     vc->virtual_server);
  else
    {
      /* Obtain a list of all users on the virtual server.  */
      vsd_get_uidgid (vc->server_root, NULL, NULL, &min_uid, &max_uid);
      
      while ((grp = fgetgrent (gr_stream)))
	if (grp->gr_gid >= min_uid && grp->gr_gid <= max_uid)
	  {
	    char **list = grp->gr_mem;
	    io_buffer_store (vc, "name=%s gid=%d members=\"",
			     grp->gr_name, grp->gr_gid);
	    while (*list)
	      {
		io_buffer_store (vc, "%s", *list);
		if (list[1])
		  io_buffer_store (vc, ",");

		list++;
	      }
	    io_buffer_store (vc, "\"\n");
	  }	    

      vsd_closegrp (gr_stream);
    }
}

/* Add a user to a virtual server.  */
void user_add (struct connection *vc, int argc, char *argv[])
{
  char line[256];
  const char *username = argv[0], *password = NULL;
  const char *homedir = NULL, *vsdir;
  uid_t uid = 0;
  gid_t gid = 0;
  int status;

  if (is_name_valid (vc, USER, vc->virtual_server, username, 0))
    return;

  vsdir = vc->server_root;
  /* Parse optional command line arguments.  */
  if (argc >= 1)
    {
      int x;

      for (x = 0; x <= argc - 1; x++)
	{
	  if (! password)
	    password = get_option ("pass", argv[x]);
	  if (! homedir)
	    homedir = get_option ("home", argv[x]);
	}
    }

  if (homedir && is_home_dir_valid (vc, vc->virtual_server, homedir))
    return;

  if (! homedir)
    {
      /* The default home directory is /home/user.  */
      sprintf (line, "/home/%s", username);
      homedir = line;
    }

  /* The default group is the same as the user name.  */
  status = vsd_groupadd (vsdir, username, &gid);
  /* Add the user to the virtual server.  */
  status = vsd_useradd (vsdir, username, homedir, username, "/bin/bash", &uid,
			NULL);
  if (status == 9)
    io_buffer_store (vc, "-ERR User is already defined");
  else
    {
      /* Create the user's home directory.  */
      if (homedir == NULL)
	sprintf (line, "%s/home/%s", vsdir, username);
      else
	sprintf (line, "%s/%s", vsdir, homedir);
      mkdir (line, (mode_t) 0755);
      chown (line, uid, gid);

      if (password != NULL
	  && vsd_chpass (vsdir, username, 1, NULL, password))
	io_buffer_store (vc, "-ERR User added but couldn't set password");
    }
}

/* Delete a user from a virtual server.  */
void user_del (struct connection *vc, int argc, char *argv[])
{
  const char *vsdir;
  int x, membc, status;
  int flags_delete_home, flags_delete_group, flags_delete_files;
  int flags_change_uids;
  struct vsd_file_line *priv_file;
  char **membv;

  if (is_name_valid (vc, USER, vc->virtual_server, argv[0],
		     NAME_DONT_CHECK_LEN))
    return;
  if (strcmp (argv[0], "admin") == 0 || strcmp (argv[0], "web") == 0
      || strcmp (argv[0], "ftp") == 0 || strcmp (argv[0], "mail") == 0)
    {
      io_buffer_store (vc, "-ERR Deleting the %s user is silly");
      return;
    }

  vsdir = vc->server_root;
  flags_delete_home = flags_delete_group = 0;
  flags_delete_files = flags_change_uids = 0;
  flags_delete_group = 1;
  for (x = 1; x < argc; x++)
    {
      if (strcmp (argv[x], "-r") == 0)
	flags_delete_home = 1; /* Delete the user's home directory.  */
      if (strcmp (argv[x], "-g") == 0)
	flags_delete_group = 1; /* Delete the user's associated group.  */
      if (strcmp (argv[x], "-f") == 0)
	flags_delete_files = 1; /* Delete all files belonging to the user.  */
      if (strcmp (argv[x], "-c") == 0)
	flags_change_uids = 1; /* Make all user's files owned by admin.  */
    }

  /* Does the user already exist ? */
  status = vsd_userdel (vsdir, argv[0]);
  if (status == 6)
    {
      io_buffer_store (vc, "-ERR User %s does not exist", argv[0]);
      return;
    }

  /* Delete the privileges for this user.  */
  if (vsd_priv_load (&priv_file, vc->server_root))
    {
      io_buffer_store (vc, "-ERR Could not delete privileges for user: %s",
		       strerror (errno));

      /* Limp on and attempt to delete the user's group, if requested.  */
    }
  else
    {
      /* Get a list of privileges that `old_user' is a member of.  */
      membv = vsd_priv_get_members (priv_file, argv[0], &membc);

      /* Delete `old_user' from the privileges and add `new_user' to
	 the privilegs.  */
      for (x = 0; x < membc; x++)
	vsd_priv_delete_member (priv_file, membv[x], argv[0]);

      vsd_priv_save (priv_file, vc->server_root);
      vsd_priv_free (priv_file);
      vsd_argv_free (membv, membc);
    }

  if (flags_delete_group)
    {
      status = vsd_groupdel (vsdir, argv[0]);
      if (status == 6)
	io_buffer_store (vc, "-ERR Group %s does not exist", argv[0]);
    }
}

/* Modify a user.  */
void user_mod (struct connection *vc, int argc, char *argv[])
{
  const char *username, *password = NULL, *homedir = NULL;
  const char *shell = NULL, *new_username = NULL, *vsdir, *gecos = NULL;
  char *p, *cmdline;
  int status;

  /* Command syntax explanation:
       --pass=<password> - new password for the user
       --home=<home dir> - the user's new login directory
       --shell=<default shell> - the name of the user's new login shell
       --login=<login-name> - change the user's account name to <login-name>
  */
  username = argv[0];
  if (is_name_valid (vc, USER, vc->virtual_server, username, 0))
    return;

  /* Does the user already exist ? */
  if (user_exists (vc, vc->virtual_server, username, NULL))
    {
      io_buffer_store (vc, "-ERR User %s does not exist", username);
      return;
    }

  vsdir = vc->server_root;

  /* Parse optional command line arguments.  */
  if (argc >= 1)
    {
      int x;

      for (x = 0; x <= argc - 1; x++)
	{
	  if (! password)
	    password = get_option ("pass", argv[x]);
	  if (! homedir)
	    homedir = get_option ("home", argv[x]);
	  if (! shell)
	    shell = get_option ("shell", argv[x]);
	  if (! new_username)
	    new_username = get_option ("login", argv[x]);
	  if (! gecos)
	    gecos = get_option ("gecos", argv[x]);
	}
    }

  if (homedir && is_home_dir_valid (vc, vc->virtual_server, homedir))
    return;

  if (new_username
      && is_name_valid (vc, USER, vc->virtual_server, username, 0))
    return;


  status = vsd_usermod (vsdir, username, new_username, homedir,
			NULL /*username */, shell, -1, gecos);
  if (status == 10)
    io_buffer_store (vc, "-ERR Couldn't access /etc/passwd: %s",
		     strerror (errno));
  else if (status == 9)
    io_buffer_store (vc, "-ERR User %s is already in use", username);
  else if (status == 7)
    io_buffer_store (vc, "-ERR User %s does not exist", username);
  else if (status == 6)
   io_buffer_store (vc, "-ERR Group %s does not exist", username);
  /*   else if (status == 4)
     io_buffer_store (vc, "-ERR Uid %d is already in use", uid); */
  else if (status == 2)
    io_buffer_store (vc, "-ERR There is an error somewhere...");

  if (new_username)
    vsd_priv_rename_user (vsdir, username, new_username);

#if 0
  if (homedir)
    {
      char line[256];
      /* Create the new user home directory.  */
      if (snprintf (line, sizeof (line), "%s/%s", vsdir, homedir) < sizeof (line))
	{
	  mkdir (line, (mode_t) 0755);
	  chown (line, uid, gid);
	}
    }
#endif

  if (password != NULL
      && vsd_chpass (vc->server_root, username, 1, NULL, password))
    io_buffer_store (vc, "-ERR User modified but password not set");
}

/* Change password for a user on a virtual server.  */
void user_chpass (struct connection *vc, int argc, char *argv[])
{
  const char *user, *old_pass, *new_pass;
  int status;

  /* Format:
       CHUSERPASS <username> <old password> <new password>

     where
       <username> - user on a virtual server
       <old password> - current password for the user
       <new password> - new password for the user
  */
  user = argv[0];
  old_pass = argv[1];
  new_pass = argv[2];

  if (is_name_valid (vc, USER, vc->virtual_server, user, 0))
    return;

  /* Does the user already exist ? */
  status = vsd_chpass (vc->server_root, user, 0, old_pass, new_pass);
  if (status == 6)
    io_buffer_store (vc, "-ERR User %s does not exist", user);
  else if (status == 4)
    io_buffer_store (vc, "-ERR Incorrect password supplied for user %s", user);
  else if (status != 0)
    io_buffer_store (vc, "-ERR Password change failed with status %d", status);
}

/* Add a new group.  */
void user_group_add (struct connection *vc, int argc, char *argv[])
{
  /* Check command line validity and arguments.  */
  if (! is_name_valid (vc, GROUP, vc->virtual_server, argv[0], 0))
    {
      gid_t gid = 0;
      int status = vsd_groupadd (vc->server_root, argv[0], &gid);
      if (status == 9)
	io_buffer_store (vc, "-ERR Group %s is already defined", argv[0]);
    }
}

/* Delete an existing group.  */
void user_group_del (struct connection *vc, int argc, char *argv[])
{
  if (is_name_valid (vc, GROUP, vc->virtual_server, argv[0], 0))
    return;

  if (strcmp (argv[0], "admin") == 0 || strcmp (argv[0], "web") == 0
      || strcmp (argv[0], "ftp") == 0)
    io_buffer_store (vc, "-ERR Deleting the %s group is silly");
  else
    {
      int status = vsd_groupdel (vc->server_root, argv[0]);
      if (status == 9)
	io_buffer_store (vc, "-ERR Group %s does not exist", argv[0]);
    }
}

/* Modify an existing group.  */
void user_group_mod (struct connection *vc, int argc, char *argv[])
{
}

