/* Control virtual server operation.
   Copyright (c) 1999, 2000 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>

#include "vsd.h"
#include "vs.h"
#include "module.h"
#include "vsd.h"
#include "libvsd.h"

static void send_output (struct connection *vc,
			 const char *server_root, const char *file)
{
  char *temp;
  FILE *stream;
  char line[256];

  if (server_root)
    {
      temp = (char *) malloc (strlen (server_root) + strlen (file) + 3);
      strcpy (temp, server_root);
      strcat (temp, file);
    }
  else
    temp = (char *) file;

  stream = fopen (temp, "r");
  if (stream != NULL)
    {
      while (fgets (line, sizeof (line), stream))
	io_buffer_store (vc, "%s", line);

      fclose (stream);
    }
  unlink (temp);

  if (server_root)
    free (temp);
}

void vs_reboot (struct connection *vc, int argc, char *argv[])
{
  char line[256];
  int status;
  char server_temp[128];
  const char *vsdir;

  vsdir = vsd_map_server_root (vc->map, vc->virtual_server);

  /* Stop the virtual server (switches to runlevel 0).
     Send all output to a file. */
  sprintf (server_temp, "/tmp/vsd.reboot.%s", vc->virtual_server);
  sprintf (line, "--restart %s", vc->virtual_server);
  status = run_command (vc, NULL, NULL, "/usr/sbin/vsboot", line, NULL,
			server_temp);

  /* Open file and send all data back to the client.  */
  send_output (vc, NULL, server_temp);
}

void vs_signal (struct connection *vc, int argc, char *argv[])
{
  int *procs = vsd_enum_procs (vc->virtual_server);
  int x = 0;
  int signum = atoi (argv[0]);

  while (procs[x] != -1)
    {
      kill (procs[x], signum);
      x++;
    }
  free (procs);
}  

void vs_account_enable (struct connection *vc, int argc, char *argv[])
{
  char line[256], server_temp[128];
  struct vsd_vs *vs;

  if (is_name_valid (vc, SERVER, argv[0], NULL, 0))
    return;

  /* Check virtual server exists.  */
  vs = vsd_getmapent (vc->map, argv[0], NULL);
  if (vs == NULL)
    {
      io_buffer_store (vc, "-ERR Server %s is not known", argv[0]);
      return;
    }

  /* Activate the account.  */
  if (vsd_map_set_status (vc->map, argv[0], "active"))
    {
      io_buffer_store (vc, "-ERR Could not find server %s: %s", argv[0],
		       strerror (errno));
      return;
    }

  vsd_map_save (vc->map);

  /* Start the virtual server and assign its ethernet address. */
  sprintf (line, "--start %s", argv[0]);
  sprintf (server_temp, "/tmp/vsd.enable.%s", argv[0]);
  if (! run_command (vc, NULL, NULL, "/usr/sbin/vsboot", line, NULL,
		     server_temp))
    {
      /* Open file and send all data back to the client.  */
      send_output (vc, NULL, server_temp);
      /* Send a SIGHUP to the inetd super server.  */
      run_command (vc, NULL, NULL, "/usr/bin/killall", "-HUP inetd", NULL,
		   NULL);
    }
}

void vs_account_disable (struct connection *vc, int argc, char *argv[])
{
  char line[256], server_temp[128];
  struct vsd_vs *vs;

  if (is_name_valid (vc, SERVER, argv[0], NULL, 0))
    return;

  /* Check virtual server exists.  */
  vs = vsd_getmapent (vc->map, argv[0], NULL);
  if (vs == NULL)
    {
      io_buffer_store (vc, "-ERR Server %s is not known", argv[0]);
      return;
    }

  /* Stop the virtual server and deassign it's ethernet address. Send all
     output to a file. */
  sprintf (line, "--stop %s", argv[0]);
  sprintf (server_temp, "/tmp/vsd.disable.%s", argv[0]);
  if (run_command (vc, NULL, NULL, "/usr/sbin/vsboot", line, NULL, server_temp))
    return;

  /* Disable the account.  */
  if (vsd_map_set_status (vc->map, argv[0], "disabled"))
    io_buffer_store (vc, "-ERR Could not find server %s: %s", argv[0],
		     strerror (errno));
  else
    {
      vsd_map_save (vc->map);
      /* Open file and send all data back to the client.  */
      send_output (vc, NULL, server_temp);
      /* Send a SIGHUP to the inetd super server.  */
      run_command (vc, NULL, NULL, "/usr/bin/killall", "-HUP inetd", NULL,
		   NULL);
    }
}

void vs_account_create (struct connection *vc, int argc, char *argv[])
{
  FILE *stream;
  struct vsd_vs *vs;
  int num_uid, start_uid, diskquota;

  /* VS_CREATE <virtual server> <ip-addr> <fqdn> <numuid> <quota> */

  /* Check validity of virtual server name.  */
  if (is_name_valid (vc, SERVER, argv[0], NULL, 0))
    return;

  /* Does virtual server exist ?  */
  vs = vsd_getmapent (vc->map, argv[0], NULL);
  if (vs != NULL)
    {
      io_buffer_store (vc, "-ERR A server of name %s already exists",
		       argv[0]);
      return;
    }

  vs = vsd_getmapent (vc->map, NULL, argv[1]);
  if (vs != NULL)
    {
      io_buffer_store (vc, "-ERR IP address %s is already in use",
		       argv[1]);
      return;
    }

  /* Find a free UID allocation.  */
  num_uid = atoi (argv[3]);
  start_uid = vsd_alloc_uid (vc->map, num_uid);
  if (start_uid == -1)
    {
      io_buffer_store (vc, "-ERR Could not allocate a UID chunk");
      return;
    }

  diskquota = atoi (argv[4]);
  if (vsd_map_add_server (vc->map, argv[0], argv[1], start_uid, num_uid,
			  diskquota))
    io_buffer_store (vc, "-ERR Cannot create account: %s",
		     (errno == EMFILE) ? "host server is full"
		     : strerror (errno));
  else
    {
      vsd_map_save (vc->map);

      stream = fopen ("/tmp/vsd.accounts", "a");
      fprintf (stream, "create %s %s\n", argv[0], argv[2]);
      fclose (stream);
    }
}

void vs_account_delete (struct connection *vc, int argc, char *argv[])
{
  FILE *stream;
  struct vsd_vs *vs;
  char line[128];
  
  if (is_name_valid (vc, SERVER, argv[0], NULL, 0))
    return;

  vs = vsd_getmapent (vc->map, argv[0], NULL);
  if (vs == NULL)
    {
      io_buffer_store (vc, "-ERR Server %s is not known", argv[0]);
      return;
    }

  /* Stop the virtual server and deassign it's ethernet address. Send all
     output to a file. */
  sprintf (line, "--stop %s", argv[0]);
  if (run_command (vc, NULL, NULL, "/usr/sbin/vsboot", line, NULL,
		   "/dev/null"))
    return;

  stream = fopen ("/tmp/vsd.accounts", "a");
  fprintf (stream, "delete %s %s\n", argv[0],
	   vc->map->partition[vs->partition].pathname);

  fclose (stream);
  vsd_map_delete_server (vc->map, argv[0]);
  vsd_map_save (vc->map);
}

void vs_details (struct connection *vc, int argc, char *argv[])
{
  struct vsd_vs *vs;
  int x, y;

  if (argc == 1)
    {
      /* Check validity of virtual server name.  */
      if (is_name_valid (vc, SERVER, argv[0], NULL, 0))
	return;

      vs = vsd_getmapent (vc->map, argv[0], NULL);
      if (vs == NULL)
	io_buffer_store (vc, "-ERR Server %s is not known", argv[0]);
      else
	{
	  io_buffer_store (vc,
			   "name=%s ip=%s uid=%d users=%d quota=%d path=\"%s\"",
			   vs->name, vs->ip,
			   vs->startuid, vs->users, vs->quota,
			   vc->map->partition[vs->partition].pathname);
	  for (x = 0; x < vc->map->ipaliases; x++)
	    if (strcmp (vc->map->ipalias[x].name, argv[0]) == 0)
	      io_buffer_store (vc, " %s", vc->map->ipalias[x].ip);
	  io_buffer_store (vc, "\n");
	}
    }
  else
    for (x = 0; x < vc->map->servers; x++)
      {
	vs = &vc->map->server[x];
	if (vs->name[0] != '\0')
	  {
	    io_buffer_store (vc,
			     "name=%s ip=%s uid=%d users=%d quota=%d path=\"%s\"",
			     vs->name, vs->ip,
			     vs->startuid, vs->users, vs->quota,
			     vc->map->partition[vs->partition].pathname);
	    for (y = 0; y < vc->map->ipaliases; y++)
	      if (strcmp (vc->map->ipalias[y].name, vs->name) == 0)
		io_buffer_store (vc, " %s", vc->map->ipalias[y].ip);
	    io_buffer_store (vc, "\n");
	  }
      }
}


