/* Useful utility functions common to various modules.
   Copyright (c) 1999, 2000 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <ctype.h>
#include <sys/syslog.h>

#include "vsd.h"
#include "libvsd.h"

#define MAX_USERNAME_LENGTH 16

const char *get_option (const char *option, const char *cmdline)
{
  char opt[32];
  const char *value;

  value = NULL;

  sprintf (opt, "--%s=", option);
  if (strncmp (cmdline, opt, strlen (opt)) == 0)
    {
      value = cmdline + strlen (opt);
      if (*value == '\0')
	value = NULL;
    }

  return value;
}

/* Return the maximum length a user name can be on virtual server
   `server_name'.  */
int max_username_len (const char *server_name)
{
  if (server_name == NULL)
    return MAX_USERNAME_LENGTH;

  return MAX_USERNAME_LENGTH - (strlen (server_name) + 1);
}

/* Return 0 if the username is valid
   Return 1 if the username is too long
   Return 2 if it contains invalid characters
   Return -1 if the argument list is invalid.  */ 
int is_name_valid (struct connection *vc, int user,
		   const char *server_name, const char *username, int options)
{
  char temp[20];

  if (user == USER)
    strcpy (temp, "User");
  else if (user == SERVER)
    {
      username = server_name;
      server_name = NULL;
      strcpy (temp, "Virtual server");
    }
  else
    strcpy (temp, "Group");

  if (username == NULL)
    {
      /* Parameters are invalid.  */
      errno = EINVAL;
      return -1;
    }

  if (strspn (username, username_reject_chars) != 0)
    {
      io_buffer_store (vc, "-ERR %s name %s contains unsuitable characters",
	      temp, username);
      return 2; /* User name contains unsuitable characters.  */
    }

  if (isdigit (username[0]))
    {
      io_buffer_store (vc, "-ERR %s name %s cannot start with a digit",
	      temp, username);
      return 2;
    }

  if (user != SERVER)
    /* The shadow password suite restricts us to 16 characters.  */
    if (! (options & NAME_DONT_CHECK_LEN) &&
	strlen (username) > MAX_USERNAME_LENGTH)
      {
	syslog (LOG_WARNING, "%s name %s is too long", temp, username);
	io_buffer_store (vc, "-ERR %s name is restricted to %d characters",
			 temp, max_username_len (NULL));
	return 1;
      }

  return 0;
}

int is_home_dir_valid (struct connection *vc, const char *server_name,
		       const char *home_dir)
{
  if (home_dir == NULL)
    {
      errno = EINVAL;
      return -1;
    }

  if (strspn (home_dir, pathname_reject_chars) != 0)
    {
      io_buffer_store (vc, "-ERR Path for home directory contains invalid characters");
      return 1;
    }

  return 0;
}

/* Return 0 if user exists on the virtual server, or 1 if it doesn't.  */
int user_exists (struct connection *vc, const char *server_name,
		 const char *user_name, uid_t *uid)
{
  FILE *stream;
  struct passwd *pw = NULL;

  stream = vsd_openpw (vc, server_name);
  if (stream)
    {
      pw = vsd_getpwnam (stream, user_name);
      vsd_closepw (stream);
    }

  if (pw == NULL)
    return 1;

  if (uid != NULL)
    *uid = pw->pw_uid;

  return 0;
}

/* Return 0 if group exists on the virtual server, or 1 if it doesn't.  */
int group_exists (struct connection *vc, const char *server_name,
		  const char *group_name)
{
  FILE *stream;
  struct group *grp = NULL;

  stream = vsd_opengrp (vsd_map_server_root (vc->map, server_name));
  if (stream)
    {
      grp = vsd_getgrnam (stream, group_name);
      vsd_closegrp (stream);
    }
  return (grp == NULL) ? 1 : 0;
}

