/* Client frontend for the virtual server administrator.
   Copyright (c) 1999, 2000 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdarg.h>

#include "module.h"
#include "vsd_lib.h"
#include "libvsd.h"

#undef TEST_MODE

char *program_name;

static void error (int usage,
		   const char *fmt, ...) __attribute__ ((__noreturn__));

static void error (int usage, const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  if (usage == 1)
    fprintf (stderr, "usage: %s ", program_name);
  else
    fprintf (stderr, "%s: ", program_name);

  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);

  exit (1);
}

int main (int argc, char *argv[])
{
  char *s, **vec;
  vsd_server *server;
  char host_server[256], virtual_server[256];
  struct passwd *pwd = NULL; /* Stop gcc complaining.  */
  int status, x;
  const char *ret;

  vsd_set_interface (CLI_INTERFACE);

  program_name = strrchr (argv[0], '/') + 1;
  if (program_name == (void *)1)
    program_name = argv[0]; /* Program wasn't called with a path.  */

  ret = module_syntax (ADMIN_CLIENT, program_name, argc - 1, argv + 1);
  if (ret != NULL)
    {
      if (strcmp (ret, "notexist") == 0)
	error (0, "%s is not a recognised executable", program_name);
      else
	error (1, "%s", ret);
    }

  /* Get the host server and virtual server names.  */
  if (vsd_gethostservername (host_server, sizeof (host_server) - 1))
    error (0, "couldn't get host server name: %s", strerror (errno));

#ifdef TEST_MODE
  strcpy (virtual_server, "test");
#else
  if (vsd_getvirtualservername (virtual_server,
				sizeof (virtual_server) - 1))
    error (0, "couldn't get virtual server name: %s",
	   strerror (errno));
#endif
  s = strchr (host_server, '.');
  if (s == NULL)
    {
      if (strcmp (host_server, virtual_server) == 0)
	error (0, "must be run on the virtual server");
    }
#if 0
  else
    {
      if (strncmp (virtual_server, host_server, s - host_server) == 0)
	error (0, "must be run on the virtual server");
    }
#endif

#if 1
  /* Verify that we are the 'admin' user.  */
  pwd = getpwuid (getuid ());
  if (pwd == NULL
      || (strcmp (pwd->pw_name, "admin") != 0
	  && strcmp (pwd->pw_name, "root") != 0))
    error (0, "must be admin user to run this program");
#endif

  /* Connect to the virtual server adminstrator.  */
  server = vsd_open (host_server, virtual_server, "admin", "NOAUTH");
  if (server == NULL)
    error (0, "connection failed: %s", vsd_geterror ());

  /* Perform the action on the virtual server.  */

  /* Align argument vector into the same format as is passed by
     webclients and webadmclients.  */
  vec = malloc (sizeof (char *) * (argc + 1));
  
  vec[0] = host_server;
  vec[1] = virtual_server;
  for (x = 1; x < argc; x++)
    vec[x + 1] = argv[x];
  
  s = module_command (server, ADMIN_CLIENT, program_name, vec);
  free (vec);

  if (s && strncmp (s, "-ERR", 4) == 0)
    {
      fprintf (stdout, "error: %s\n", s);
      vsd_close (server);
      return 1;
    }

  fprintf (stdout, "%s", s);
  vsd_close (server);

  return 0;
}

