/* Virtual server administrator.
   Copyright (c) 1999 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "config.h"

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#else
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdarg.h>
#include <time.h>
#include <netdb.h>
#include <grp.h>
#include <pwd.h>
#include <fcntl.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/time.h>
#include <sys/resource.h>

#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#include "vsd.h"
#include "module.h"
#include "libvsd.h"

/* Autologout timer */
#define LOGINTIMEOUT 60*3
#define TIMEOUT 60*10

/* List of characters that we do not allow in a username.  */
const char *username_reject_chars="!\"$%^&*()+=}{~#@':;?>.<,?/|\\ ";
/* List of characters that we do not allow in a pathname.  */
const char *pathname_reject_chars="!\"$%^&*()}]{[?<>:;\\ ";


/* Function prototypes */

/* getdelim doesn't seem to be declared in stdio.h.  */
extern ssize_t getdelim (char **lineptr, size_t *n, int delim, FILE *stream);
static void process_requests (struct connection *vc);
static void free_connection (struct connection *vc);

/* Clock interrupt.  */
static void clkint (int sig)
{
  output (NULL, "-ERR Autologout; idle for too long\n");
  /* syslog (LOG_INFO, "Autologout user=%.80s host=%.80s",
     username ? username : "???", tcp_clienthost ()); */
  _exit (1);
}

/* Hangup interrupt.  */

static void hupint (int sig)
{
  alarm (0);
  /* syslog (LOG_INFO, "Hangup user=%.80s host=%.80s",
     username ? username : "???", tcp_clienthost ()); */

  _exit (1);
}


/* Termination interrupt.  */

static void trmint (int sig)
{
  alarm (0);

  output (NULL, "-ERR Killed\n");
  /* syslog (LOG_INFO, "Killed user=%.80s host=%.80s",
     username ? username : "???", tcp_clienthost ()); */

  _exit (1);
}

static char *localhost (void)
{
  static char *local_host = NULL;

  if (local_host == NULL)
    {
      char tmp[1024];
      struct hostent *host_name;

      gethostname (tmp, sizeof (tmp));
      host_name = gethostbyname (tmp);
      if (host_name)
	local_host = strdup (host_name->h_name);
      else
	local_host = strdup (tmp);
    }

  return local_host;
}

char *tcp_serverhost (void)
{
  static char *server_host = NULL;
  static long server_port = -1;

  if (server_host == NULL)
    {
      char tmp[1024];
      struct hostent *he;
      struct sockaddr_in sin;
      int sin_len = sizeof (struct sockaddr_in);

      if (getsockname (0, (struct sockaddr *) &sin, (void *) &sin_len))
	strcpy (tmp, localhost ());
      else
	{
	  server_port = ntohs (sin.sin_port);
	  he = gethostbyaddr ((char *) &sin.sin_addr,
			      sizeof (struct in_addr), sin.sin_family);
	  if (he)
	    strcpy (tmp, he->h_name);
	  else
	    sprintf (tmp, "[%s]", inet_ntoa (sin.sin_addr));
	}
      server_host = strdup (tmp);
    }

  return server_host;
}

void tcp_clienthost (struct connection *vc)
{
  char tmp[1024];
  struct hostent *he;
  struct sockaddr_in sin;
  int sin_len = sizeof (struct sockaddr_in);

  if (getpeername (0, (struct sockaddr *) &sin, (void *) &sin_len))
    strcpy (tmp, "Unknown");
  else
    {
      he = gethostbyaddr ((char *) &sin.sin_addr,
			  sizeof (struct in_addr), sin.sin_family);
      if (he)
	sprintf (tmp, "%s [%s]", he->h_name, inet_ntoa (sin.sin_addr));
      else
	sprintf (tmp, "[%s]", inet_ntoa (sin.sin_addr));
    }
  
  vc->client_host = strdup (tmp);
}

#ifndef WITH_OPENSSL
static void cleanup (void)
{
}
#endif

#ifdef WITH_OPENSSL
static void cleanup (void)
{
  SSL_free (ssl_ssl);
  SSL_CTX_free (ssl_ctx);
}

#endif
/* Output a string and flush it.  */
void output (struct connection *vc, const char *fmt, ...)
{
  va_list ap;
  char *s = NULL;
  int len;

  va_start (ap, fmt);
  len = vsnprintf (s, 0, fmt, ap);
  s = (char *) malloc (len + 1);
  vsprintf (s, fmt, ap);

  while (write (fileno (stdout), s, len) == -1 && errno == EINTR)
      ;
  free (s);
  va_end (ap);
}

/* Main program */
int 
main (int argc, char *argv[])
{
  char *s;
  struct sigaction nact, oact;
  struct connection *vsd_connection;

  /* If the program name doesn't exist, then create our own for the
     syslog daemon.  */
  s = argv[0];
  if (s == NULL)
    s = "vsd";
  openlog (s, LOG_PID, LOG_DAEMON);

#if 0
  /* Check we are running as root.  */
  if (getuid () != 0)
    {
      fprintf (stderr, "%s: must be run as root\n", s);
      syslog (LOG_ERR, "must be run as root");
      return 1;
    }
#endif

  vsd_connection = (struct connection *) malloc (sizeof (struct connection));
  memset (vsd_connection, 0, sizeof (struct connection));
  tcp_clienthost (vsd_connection);

  /* Set a 10 second timeout on command input.  */
  vsd_connection->command_timeout = 30;

#if 0
  /* Authenticate server access.  */
  if (authenticate_server (vsd_connection) == 0)
    {
      io_buffer_store (vc, "-ERR Connection refused\n");
      return 1;
    }
#endif

  sigemptyset (&nact.sa_mask);
  nact.sa_handler = clkint;
  nact.sa_flags = SA_RESTART;
  sigaction (SIGALRM, &nact, &oact);
  nact.sa_handler = hupint;
  sigaction (SIGHUP, &nact, &oact);
  nact.sa_handler = trmint;
  sigaction (SIGTERM, &nact, &oact);

  vsd_connection->map = vsd_map_read ();
  if (vsd_connection->map == NULL)
    output (vsd_connection, "-ERR Fatal error: server map read failure: %s\n",
	    strerror (errno));
  else
    {
      output (vsd_connection,
	      "+OK %s Virtual Server administrator %s v%s server ready\n",
	      PACKAGE, tcp_serverhost (), VERSION);
      /* Retrieve incoming data, parse and store in a io_request structure.  */
      io_parse_input (vsd_connection);

      /* Process the requests in the io_request structure and create an
	 io_output structure containing the replies.  */
      process_requests (vsd_connection);
      /* Send the data back to the client.  */
      io_generate_output (vsd_connection);

      output (vsd_connection, "+OK Virtual server admin signing off\n");
    }

  /* Free all memory used by the structures.  */
  free_connection (vsd_connection);
  cleanup ();

  return 0;
}

/* Free memory allocated within the connection structure.  */
static void free_connection (struct connection *vc)
{
  if (vc->input_buffer)
    free (vc->input_buffer);
  if (vc->virtual_server)
    free (vc->virtual_server);
  if (vc->user_name)
    free (vc->user_name);
  if (vc->server_root)
    free (vc->server_root);
  if (vc->client_host)
    free (vc->client_host);
  io_buffer_free (vc);
}

static void process_requests (struct connection *vc)
{
  int req, cmd, level;
  const char *status;

  for (req = 0; req < vc->requests_count; req++)
    {
      vc->virtual_server = vc->requests[req].vs;
      vc->vs = vsd_getmapent (vc->map, vc->virtual_server, NULL);
      /* Some commands are not appropriate for a specific virtual server.
	 These are entered as a VS called `<root>'.  */
      status = vsd_map_server_root (vc->map, vc->virtual_server);
      vc->server_root = (status) ? strdup (status) : NULL;
      level = (vc->server_root) ? ADMIN_VS : ADMIN_HOST;
      for (cmd = 0; cmd < vc->requests[req].actionc; cmd++)
	{
	  status = module_syntax (level,
				  vc->requests[req].actionv[cmd].argv[0],
				  vc->requests[req].actionv[cmd].argc - 1,
				  vc->requests[req].actionv[cmd].argv + 1);
	  if (status && ! strcmp (status, "notexist")) /* No such command.  */
	    io_buffer_store (vc, "No such command");
	  if (status && strcmp (status, "notexist")) /* Syntax error.  */
	    io_buffer_store (vc, "-ERR Syntax: %s", status);
	  else if (! status) /* Command exists.  */
	    module_transaction (vc, level,
				vc->requests[req].actionv[cmd].argv[0],
				vc->requests[req].actionv[cmd].argc - 1,
				vc->requests[req].actionv[cmd].argv + 1,
				vc->requests[req].actionv[cmd].data);
	  io_collect_output (vc,
			     vc->requests[req].actionv[cmd].argv[0]);
	}
      free (vc->server_root);
    }
}

/* Fork and execute `program_name' as a child process as the user `username'
   and wait until completion. Optionally `parameters' can contain a space
   separated string of command line arguments for the process.  The child
   process will have it's root directory set to `chroot_path' if not
   NULL.  Optionally stdin, stdout and stderr can be redirected to
   the filenames specified by `redir_stdin', `redir_stdout'.

   Return 0 on success, any other value on error.  */
int run_command (struct connection *vc,
		 const char *chroot_path, const char *username,
		 const char *program_name, const char *parameters,
		 const char *redir_stdin, const char *redir_stdout)
{
  int status, len;
  pid_t pid;
  char *line;

  if (program_name == NULL)
    {
      io_buffer_store (vc, "-ERR run_command: no program name");
      return -1;
    }

  len = strlen (program_name);
  if (len < sizeof ("/bin/sh") + 1)
    len = sizeof ("/bin/sh") + 1;
  if (chroot_path)
    len += strlen (chroot_path);
  len += 2; /* Cope with spaces and terminators.  */
  line = (char *) malloc (len);

  /* Check that `program_name' exists and is executable.  If we do this
     now then we can output more accurate error messages back to the
     client.  These checks should substantially reduce the risk of the
     client applications failing.  */
  line[0] = '\0';
  if (chroot_path)
    strcpy (line, chroot_path);
  strcat (line, program_name);

  if (access (line, R_OK | X_OK) == -1)
    {
      /* Don't give full pathname to the user. Only give local pathname
	 w.r.t. their virtual server.  */
      io_buffer_store (vc, "-ERR Access check on %s failed: %s",
	      program_name, strerror (errno));

      /* Log as an error with syslogd with full pathname to aid the
	 sysadmin.  */
      syslog (LOG_ERR, "run_command: access check on %s failed: %m",
	      line);
      return -1;
    }

  /* Check that /bin/sh exists and is executable.  */
  line[0] = '\0';
  if (chroot_path)
    strcpy (line, chroot_path);
  strcat (line, "/bin/sh");

  if (access (line, R_OK | X_OK) == -1)
    {
      io_buffer_store (vc, "-ERR Access check on /bin/sh failed: %s",
	      strerror (errno));

      syslog (LOG_ERR, "run_command: access check on %s failed: %m",
	      line);
      return -1;
    }

  free (line);

  pid = fork ();
  if (pid == (pid_t) 0)
    {
      /* Child side.  */
      const char *new_argv[4];
      const char *new_environ[2];
      char *shell_arg;
      int in_fd, out_fd;
#if 0
      struct rlimit lim;
#endif

      shell_arg = (char *) malloc (strlen (program_name) + strlen (parameters)
				   + 3);
      sprintf (shell_arg, "%s %s", program_name, parameters);

      new_argv[0] = "sh";
      new_argv[1] = "-c";
      new_argv[2] = shell_arg;
      new_argv[3] = NULL;

      new_environ[0] = "PATH=/usr/local/bin:/bin:/usr/bin";
      new_environ[1] = NULL;

#if 0
      /* Set CPU time and Memory usage limits. We don't want run away
	 processes fucking over our system. There's a bug in useradd
	 that consumes all CPU and memory when it comes across an incorrect
	 entry in /etc/group.  */

      /* Limit CPU time to 4 seconds.  */
      lim.rlim_cur = 4;
      lim.rlim_max = 4;
      setrlimit (RLIMIT_CPU, &lim);

      /* Limit number of processes to 2.  */
      lim.rlim_cur = 4;
      lim.rlim_max = 4;
      setrlimit (RLIMIT_NPROC, &lim);

      /* Limit memory to 16Mb.  */
      lim.rlim_cur = 16*1024*1024;
      lim.rlim_max = 16*1024*1024;
      setrlimit (RLIMIT_DATA, &lim);
#endif

      if (chroot_path != NULL)
	{
	  if (chroot (chroot_path) == -1)
	    {
	      io_buffer_store (vc, "-ERR chroot failed");
	      syslog (LOG_ERR, "chroot (%s) failed: %m", chroot_path);
	      _exit (1);
	    }

	  if (chdir ("/") != 0)
	    {
	      io_buffer_store (vc, "-ERR chdir failed");
	      syslog (LOG_ERR, "chdir failed: %m");
	      _exit (1);
	    }
	}

      if (username != NULL)
	{
	  /* Become user `username'.  */
	  struct passwd *pw;

	  pw = getpwnam (username);
	  setuid (pw->pw_uid);
	}

      if (redir_stdin != NULL)
	{
	  /* Take standard input from `redir_stdin'.  */
	  close (0);
	  in_fd = open (redir_stdin, O_RDONLY);
	}

      if (redir_stdout != NULL)
	{
	  /* Redirect child process output to a file.  */
	  close (1);
	  close (2);
	  out_fd = open (redir_stdout, O_CREAT | O_WRONLY, 0600);
	  fcntl (out_fd, F_DUPFD, 2);
	}

      /* Exec the shell.  */
      execve ("/bin/sh", (char *const *) new_argv, (char *const *) new_environ);
      io_buffer_store (vc, "-ERR cannot execute %s", new_argv[0]);

      syslog (LOG_ERR, "cannot execute %s: %m", new_argv[0]);
      _exit (127);
    }
  else if (pid < (pid_t) 0)
    {
      /* The fork failed.  */
      status = -1;
    }
  else
    {
      /* Parent side.  */
      if (waitpid (pid, &status, 0) != pid)
	status = -1;

      /* Did exit occur successfully ?  */
      if (WIFEXITED (status) && WEXITSTATUS (status) == 0)
	return 0;

      /* Output an error if the program terminated without success,  */
      if (WIFSIGNALED (status))
	{
	  io_buffer_store (vc, "-ERR Program %s got fatal signal %d",
		  program_name, WTERMSIG (status));

	  /* Log to syslog to aid the sysadmin.  */
	  syslog (LOG_ERR, "run_command: program %s%s got fatal signal %d",
		  (chroot_path == NULL) ? "" : chroot_path,
		  program_name, WTERMSIG (status));
	}
      else if (WIFEXITED (status)
	       && WEXITSTATUS (status) != 0)
	{
	  io_buffer_store (vc, "-ERR Program %s exited with status %d",
		  program_name, WEXITSTATUS (status));

	  syslog (LOG_ERR, "run_command: program %s%s exited with status %d",
		  (chroot_path == NULL) ? "" : chroot_path,
		  program_name, WEXITSTATUS (status));
	}
    }

  return status;
}
