/* Virtual server daemon declarations.
   Copyright (c) 1999 Designer Servers Ltd.
   Contributed by Nick Burrett <nick@dsvr.net>

   This file is part of the Virtual Server Administrator (FreeVSD)

   FreeVSD is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FreeVSD is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with FreeVSD; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#ifndef VSD_H
#define VSD_H

#include <sys/types.h>
#ifdef HAVE_LIBPAM
#include <security/pam_appl.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

struct io_requests
{
  char *vs; /* Virtual server */
  struct req_command
  {
    char **argv; /* Command and arguments for command.  */
    size_t argc; /* Number of arguments.  */
    char *data; /* Buffer for extra data that a command may pass in.  */
  } *actionv;
  size_t actionc;
};

struct io_results
{
  char *vs; /* Virtual server.  */
  struct res_command
  {
    char *cmd; /* Command name.  */
    char *data; /* Result data for command.  */
  } *cmdv;
  size_t cmdc;
};

/* Client connection structure.  */
struct connection
{
  /* Input buffer. Incoming data is placed in here.  */
  char *input_buffer;
  /* Size of the input buffer.  */
  int input_bufsiz;

  /* Name of the virtual server that the client has just logged into.  */
  char *virtual_server;
  /* User name that the client has logged in as.  */
  char *user_name;
  /* Pathname to the root of the virtual server.  */
  char *server_root;

  /* Timeouts for authentication commands and then general commands.
     Measured in seconds.  */
  int login_timeout, command_timeout;
  /* Host name of the connecting computer.  */
  char *client_host;
  /* /etc/vsd.conf.  */
  struct vsd_vs_map *map;
  /* Entry specific to the virtual server.  */
  struct vsd_vs *vs;

  /* Buffers for socket i/o.  */

  /* Output from VSD can be changed to different formats.  */
  int io_output_format;

  /* Incoming requests are parsed and stored within this data structure.  */
  struct io_requests *requests;
  int requests_count;

  /* Outgoing results are stored within this data structure.  */
  struct io_results *results;
  int results_count;

  /* Pointer to a memory chunk used as a collection point for data to be
     output.  */
  char *io_buffer;
  /* Pointer to the current position within this buffer (moves as data
     is added).  */
  char *io_bufferp;
  /* Size of the buffer.  */
  int io_buffer_size;
};

#define SERVER 2
#define USER 1
#define GROUP 0

#define NAME_DONT_CHECK_LEN 1

extern void output (struct connection *vc, const char *fmt, ...);


/* Declarations in auth.c.  */
extern int authenticate_server (struct connection *vc);
extern int login (struct connection *vc, const char *t);
extern int server_login (struct connection *vc, const char *password);
extern struct passwd *checkpw (struct passwd *pw, const char *pass);

#ifdef HAVE_LIBPAM
extern int checkpw_conv (int num_msg, const struct pam_message **msg,
			 struct pam_response **resp, void *appdata_ptr);
#endif

/* Declarations in vsd.c.  */

/* Server states.  */
#define STATE_AUTHORISATION 0
#define STATE_TRANSACTION 1
#define STATE_UPDATE 2
#define STATE_LOGOUT 3

extern const char *username_reject_chars, *pathname_reject_chars;

extern void tcp_clienthost (struct connection *vc);
extern char *tcp_serverhost (void);
extern int run_command (struct connection *vc,
			const char *chroot_path, const char *username,
			const char *program_name, const char *line,
			const char *redir_stdin, const char *redir_stdout);

/* Declarations in passwd.c.  */
extern struct passwd *vsd_getpwuid (FILE *stream, uid_t uid);
extern struct passwd *vsd_getpwnam (FILE *stream, const char *name);
extern FILE *vsd_openpw (struct connection *vc, const char *virtual_server);
extern void vsd_closepw (FILE *stream);

/* Declarations in proftpd.c.  */
extern int module_proftpd_check_chrt_priv (struct connection *vc,
					   const char *username);
extern int module_proftpd_add_chrt_priv (struct connection *vc,
					 const char *username);


/* Declarations in user.c.  */
extern void userlist (struct connection *vc, int argc, char *argv[]);
extern int useradd (struct connection *vc, int argc, char *argv[]);
extern int userdel (struct connection *vc, int argc, char *argv[]);
extern void usermod (struct connection *vc, int argc, char *argv[]);
extern int groupadd (struct connection *vc, int argc, char *argv[]);
extern int groupdel (struct connection *vc, int argc, char *argv[]);
extern void chuserpass (struct connection *vc, int argc, char *argv[]);
extern int modify_user_rights (struct connection *vc, int argc, char *argv[]);
extern int set_user_rights (struct connection *vc, int argc, char *argv[]);
extern int list_user_rights (struct connection *vc, int argc, char *argv[]);

/* Declarations in util.c */
extern const char *get_option (const char *option, const char *cmdline);
extern int max_username_len (const char *server_name);
extern int is_name_valid (struct connection *vc,
			  int user, const char *server_name,
			  const char *user_name, int options);
extern int is_home_dir_valid (struct connection *vc,
			      const char *server_name, const char *home_dir);
extern int user_exists (struct connection *vc, const char *server_name,
			const char *user_name, uid_t *uid);
extern int group_exists (struct connection *vc, const char *server_name,
			 const char *group_name);

/* io.c */
extern struct io_requests *io_parse_input (struct connection *vc);
extern void io_generate_output (struct connection *vc);
extern void io_collect_output (struct connection *vc, const char *command);
extern void io_buffer_store (struct connection *vc, const char *fmt, ...);
extern void io_buffer_free (struct connection *vc);

#ifdef __cplusplus
}
#endif

#endif
